"""The command line interface."""

from __future__ import annotations

# ruff: noqa: D103
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union

from click import Argument, Command, Context, Group, Option

from rich_click.decorators import command as _rich_command
from rich_click.decorators import group as _rich_group
from rich_click.rich_command import RichCommand, RichCommandCollection, RichGroup, RichMultiCommand
from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_help_formatter import RichHelpFormatter


if TYPE_CHECKING:
    import typer.core


__TyperGroup: Type["typer.core.TyperGroup"]
__TyperCommand: Type["typer.core.TyperCommand"]
__TyperArgument: Type["typer.core.TyperArgument"]
__TyperOption: Type["typer.core.TyperOption"]


class _PatchedTyperContext(RichContext):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        RichContext.__init__(self, *args, **kwargs)

    def make_formatter(self, error_mode: bool = False) -> RichHelpFormatter:
        """Create the Rich Help Formatter."""
        formatter = super().make_formatter(error_mode=error_mode)
        if hasattr(self.command, "rich_markup_mode") and self.command.rich_markup_mode == "rich":
            formatter.config.text_markup = "rich"
            formatter.config.text_emojis = True
        return formatter

    def get_help(self) -> str:
        """
        Format the help into a string and returns it.

        Calls :meth:`format_help` internally.
        """
        import typer.core
        from typer.models import DefaultPlaceholder

        from rich_click.rich_panel import RichCommandPanel

        if isinstance(self.command, RichGroup):
            command_panels: Dict[str, List[str]] = {}
            for cmd_name in self.command.commands:
                cmd = self.command.commands[cmd_name]
                if (
                    isinstance(cmd, (typer.core.TyperCommand, typer.core.TyperGroup))
                    and cmd.rich_help_panel is not None
                    and not isinstance(cmd.rich_help_panel, DefaultPlaceholder)
                ):
                    command_panels.setdefault(cmd.rich_help_panel, [])
                    command_panels[cmd.rich_help_panel].append(cmd_name)
            for name, commands in command_panels.items():
                self.command.add_panel(RichCommandPanel(name, commands=commands))

        return super().get_help()


def _typer_command_init(
    self: Any, *args: Any, rich_help_panel: Union[str, None] = None, rich_markup_mode: Any = None, **kwargs: Any
) -> None:
    import typer.core

    globals().setdefault("__TyperCommand", typer.core.TyperCommand)
    global __TyperCommand

    super(__TyperCommand, self).__init__(*args, **kwargs)

    super(typer.core.TyperCommand, self).__init__(*args, **kwargs)
    self.rich_markup_mode = rich_markup_mode
    self.rich_help_panel = rich_help_panel
    if not hasattr(self, "_help_option"):
        self._help_option = None


def _typer_group_init(
    self: Any,
    *args: Any,
    rich_help_panel: Union[str, None] = None,
    rich_markup_mode: Any = None,
    suggest_commands: bool = True,
    **kwargs: Any,
) -> None:
    import typer.core

    globals().setdefault("__TyperGroup", typer.core.TyperGroup)
    global __TyperGroup

    super(__TyperGroup, self).__init__(*args, **kwargs)
    self.rich_markup_mode = rich_markup_mode
    self.rich_help_panel = rich_help_panel
    self.suggest_commands = suggest_commands
    self.panels = []
    self._alias_mapping = {}
    if not hasattr(self, "_help_option"):
        self._help_option = None


def _parse_args(self, ctx: Context, args: List[str]) -> List[str]:  # type: ignore[no-untyped-def]
    # This is to handle a weird incompatibility with Type and later versions of Click
    # tldr: raise NoArgsIsHelp() doesn't work to render help text.
    if not args and self.no_args_is_help and not ctx.resilient_parsing:
        print(ctx.get_help())
        ctx.exit()
    return super(__TyperGroup, self).parse_args(ctx, args)  # type: ignore[no-any-return]


def _patch_typer_group(cls: Type[Group]) -> Type[Group]:
    cls.format_help = RichGroup.format_help  # type: ignore[assignment]
    cls.__init__ = _typer_group_init  # type: ignore[method-assign]
    cls.parse_args = _parse_args  # type: ignore[method-assign]
    cls.context_class = _PatchedTyperContext
    cls.panel = property(  # type: ignore[attr-defined]
        lambda self: self.rich_help_panel, lambda self, value: setattr(self, "rich_help_panel", value)
    )
    return cls


def _patch_typer_command(cls: Type[Command]) -> Type[Command]:
    cls.format_help = RichCommand.format_help  # type: ignore[assignment]
    cls.__init__ = _typer_command_init  # type: ignore[method-assign]
    cls.context_class = _PatchedTyperContext
    cls.panel = property(  # type: ignore[attr-defined]
        lambda self: self.rich_help_panel, lambda self, value: setattr(self, "rich_help_panel", value)
    )
    return cls


def _patch_typer_argument(cls: Type[Argument]) -> Type[Argument]:
    cls.panel = property(  # type: ignore[attr-defined]
        lambda self: self.rich_help_panel, lambda self, value: setattr(self, "rich_help_panel", value)
    )
    return cls


def _patch_typer_option(cls: Type[Option]) -> Type[Option]:
    cls.panel = property(  # type: ignore[attr-defined]
        lambda self: self.rich_help_panel, lambda self, value: setattr(self, "rich_help_panel", value)
    )
    return cls


class PatchMeta(type):
    def __init__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]) -> None:
        super().__init__(name, bases, namespace)
        if cls.__module__ in ["typer.core", "typer.main", "rich_click.patch"]:
            if name == "TyperGroup":
                _patch_typer_group(cls)  # type: ignore[arg-type]
            elif name == "TyperCommand":
                _patch_typer_command(cls)  # type: ignore[arg-type]
            elif name == "TyperOption":
                _patch_typer_option(cls)  # type: ignore[arg-type]
            elif name == "TyperArgument":
                _patch_typer_argument(cls)  # type: ignore[arg-type]


class _PatchedRichCommand(RichCommand, metaclass=PatchMeta):
    pass


class _PatchedRichMultiCommand(RichMultiCommand, _PatchedRichCommand):
    pass


class _PatchedRichCommandCollection(RichCommandCollection, _PatchedRichCommand):
    pass


class _PatchedRichGroup(RichGroup, _PatchedRichCommand):
    pass


# Options and Arguments don't need to be patched for base click CLIs.
# However, we need to intercept type.__init__ calls for Typer CLIs.


class _PatchedOption(Option, metaclass=PatchMeta):
    pass


class _PatchedArgument(Argument, metaclass=PatchMeta):
    pass


def rich_command(*args, **kwargs):  # type: ignore[no-untyped-def]
    kwargs.setdefault("cls", _PatchedRichCommand)
    kwargs["__rich_click_cli_patch"] = True
    return _rich_command(*args, **kwargs)


def rich_group(*args, **kwargs):  # type: ignore[no-untyped-def]
    kwargs.setdefault("cls", _PatchedRichGroup)
    kwargs["__rich_click_cli_patch"] = True
    return _rich_group(*args, **kwargs)


def patch(
    rich_config: Optional[RichHelpConfiguration] = None,
    *,
    patch_rich_click: bool = False,
    patch_typer: bool = False,
) -> None:
    """Patch Click internals to use rich-click types."""
    import warnings

    import rich_click._click_types_cache  # noqa: F401

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)

        import click
        import click.core

        rich_click.rich_command.OVERRIDES_GUARD = True
        click.group = rich_group
        click.command = rich_command
        click.Group = _PatchedRichGroup  # type: ignore[misc]
        click.Command = _PatchedRichCommand  # type: ignore[misc]
        click.CommandCollection = _PatchedRichCommandCollection  # type: ignore[misc]
        click.Argument = _PatchedArgument  # type: ignore[misc]
        click.Option = _PatchedOption  # type: ignore[misc]

        click.core.Group = _PatchedRichGroup  # type: ignore[misc]
        click.core.Command = _PatchedRichCommand  # type: ignore[misc]
        click.core.CommandCollection = _PatchedRichCommandCollection  # type: ignore[misc]
        click.core.Argument = _PatchedArgument  # type: ignore[misc]
        click.core.Option = _PatchedOption  # type: ignore[misc]

        if hasattr(click, "MultiCommand"):
            click.MultiCommand = _PatchedRichMultiCommand  # type: ignore[assignment,misc,unused-ignore]
        if patch_rich_click:
            rich_click.group = rich_group
            rich_click.command = rich_command
            rich_click.Group = _PatchedRichGroup  # type: ignore[misc]
            rich_click.Command = _PatchedRichCommand  # type: ignore[misc]
            rich_click.CommandCollection = _PatchedRichCommandCollection  # type: ignore[misc]
            if hasattr(click, "MultiCommand"):
                rich_click.MultiCommand = _PatchedRichMultiCommand  # type: ignore[assignment,misc,unused-ignore]

    if patch_typer:
        globals()["patch_typer"]()

    if rich_config is not None:
        rich_config.dump_to_globals()


def patch_typer(rich_config: Optional[RichHelpConfiguration] = None) -> None:
    import typer.core
    import typer.main

    if not issubclass(typer.core.TyperCommand, _PatchedRichCommand):
        globals().setdefault("__TyperCommand", typer.core.TyperCommand)

        class _PatchedTyperCommand(_PatchedRichCommand, typer.core.TyperCommand):  # type: ignore[misc]
            pass

        typer.core.TyperCommand = typer.main.TyperCommand = _patch_typer_command(_PatchedTyperCommand)  # type: ignore[assignment,attr-defined,misc]

    if not issubclass(typer.core.TyperGroup, _PatchedRichGroup):
        globals().setdefault("__TyperGroup", typer.core.TyperGroup)

        class _PatchedTyperGroup(_PatchedRichGroup, typer.core.TyperGroup):  # type: ignore[misc]
            pass

        typer.core.TyperGroup = typer.main.TyperGroup = _patch_typer_group(_PatchedTyperGroup)  # type: ignore[assignment,attr-defined,misc]

    if not issubclass(typer.core.TyperOption, _PatchedOption):
        globals().setdefault("__TyperOption", typer.core.TyperOption)

        class _PatchedTyperOption(_PatchedOption, typer.core.TyperOption):
            pass

        typer.core.TyperOption = typer.main.TyperOption = _patch_typer_option(_PatchedTyperOption)  # type: ignore[assignment,attr-defined,misc]

    if not issubclass(typer.core.TyperArgument, _PatchedArgument):
        globals().setdefault("__TyperArgument", typer.core.TyperArgument)

        class _PatchedTyperArgument(_PatchedArgument, typer.core.TyperArgument):
            pass

        typer.core.TyperArgument = typer.main.TyperArgument = _patch_typer_argument(_PatchedTyperArgument)  # type: ignore[assignment,attr-defined,misc]

    if rich_config is not None:
        rich_config.dump_to_globals()


__all__ = ["patch", "patch_typer"]



from typing import Generator

import pytest

from rich_click.rich_click_theme import COLORS, FORMATS, RichClickThemeNotFound, get_theme
from rich_click.rich_help_configuration import FromTheme, RichHelpConfiguration


def all_themes() -> Generator[str, None, None]:
    for c in COLORS:
        yield c
        for f in FORMATS:
            yield f
            yield f"{c}-{f}"


def test_all_themes_available() -> None:
    for theme in all_themes():
        get_theme(theme)
    assert True


def test_bad_theme_raises_err() -> None:
    with pytest.raises(RichClickThemeNotFound):
        get_theme("this-does-not-exist")


def test_all_themes_have_from_theme_objs_replaced_on_resolution() -> None:
    def _no_from_themes(config: RichHelpConfiguration) -> bool:
        for k in config.__dataclass_fields__:
            v = getattr(config, k)
            assert not isinstance(v, FromTheme)
        return True

    for theme in all_themes():
        cfg = RichHelpConfiguration(theme=theme)
        assert _no_from_themes(cfg)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, List, Optional, Union

import click


if TYPE_CHECKING:
    from rich.columns import Columns
    from rich.style import StyleType

    from rich_click.rich_context import RichContext
    from rich_click.rich_help_formatter import RichHelpFormatter
    from rich_click.rich_help_rendering import RichPanelRow
    from rich_click.rich_panel import RichOptionPanel


class RichParameter(click.Parameter):
    r"""
    A parameter to a command comes in two versions: they are either
    :class:`Option`\s or :class:`Argument`\s.  Other subclasses are currently
    not supported by design as some of the internals for parsing are
    intentionally not finalized.
    """

    def __init__(
        self,
        *args: Any,
        panel: Optional[Union[str, List[str]]] = None,
        help: Optional[str] = None,
        help_style: Optional["StyleType"] = None,
        **kwargs: Any,
    ):
        """Create RichParameter instance."""
        super().__init__(*args, **kwargs)
        self.panel = panel

        if help:
            help = inspect.cleandoc(help)

            deprecated = kwargs.get("deprecated")
            if deprecated:
                deprecated_message = f"(DEPRECATED: {deprecated})" if isinstance(deprecated, str) else "(DEPRECATED)"
                help = help + deprecated_message if help is not None else deprecated_message

        self.help = help

        self.help_style = help_style

    def to_info_dict(self) -> dict[str, Any]:
        """
        Gather information that could be useful for a tool generating
        user-facing documentation.

        Use :meth:`click.Context.to_info_dict` to traverse the entire
        CLI structure.
        """
        info = super().to_info_dict()
        info.setdefault("help", self.help)
        return info

    def get_rich_help(self, ctx: "RichContext", formatter: "RichHelpFormatter") -> "Columns":
        """Get the rich help text for this parameter."""
        from rich_click.rich_help_rendering import get_help_parameter

        return get_help_parameter(self, ctx, formatter)

    def get_rich_table_row(
        self,
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
        panel: Optional["RichOptionPanel"] = None,
    ) -> "RichPanelRow":
        """Create a row for the rich table corresponding with this parameter."""
        from rich_click.rich_help_rendering import get_parameter_rich_table_row

        return get_parameter_rich_table_row(self, ctx, formatter, panel)


class RichArgument(RichParameter, click.Argument):
    """
    Arguments are positional parameters to a command.  They generally
    provide fewer features than options but can have infinite ``nargs``
    and are required by default.

    All parameters are passed onwards to the constructor of :class:`Parameter`.
    """


class RichOption(RichParameter, click.Option):
    """
    Options are usually optional values on the command line and
    have some extra features that arguments don't have.

    All other parameters are passed onwards to the parameter constructor.
    """



import io
import json
import sys
from dataclasses import asdict
from importlib import reload
from typing import TYPE_CHECKING

import pytest
from click.testing import CliRunner
from inline_snapshot import snapshot
from rich.console import Console

import rich_click.rich_click as rc
from rich_click import RichContext, RichHelpConfiguration, command, group, rich_config


if sys.version_info < (3, 11):
    pass
else:
    pass


def test_basic_config_for_group() -> None:
    @group(invoke_without_command=True)
    @rich_config(help_config=RichHelpConfiguration(style_option="#111111"))
    def cli() -> None:
        pass

    @cli.command()
    @rich_config(help_config=RichHelpConfiguration(style_argument="#222222"))
    def subcommand1() -> None:
        pass

    @cli.command()
    @rich_config(help_config=dict(style_argument="#333333"))
    def subcommand2() -> None:
        pass

    @cli.command()
    @rich_config(help_config=dict(style_option="#444444"))
    def subcommand3() -> None:
        pass

    with cli.make_context("pytest-example", []) as ctx:
        if TYPE_CHECKING:
            # TODO: I don't understand this Mypy error.
            #  I wrap this one in if TYPE_CHECKING to signify that Mypy needs this.
            #  Every make_context() call in this file fails the type check though.
            assert isinstance(ctx, RichContext)
        assert ctx.help_config.style_option == "#111111"

        with subcommand1.make_context("pytest-example", [], parent=ctx) as sub_ctx:
            assert isinstance(sub_ctx, RichContext)
            assert sub_ctx.help_config.style_option == RichHelpConfiguration().style_option
            assert sub_ctx.help_config.style_argument == "#222222"

        with subcommand2.make_context("pytest-example", [], parent=ctx) as sub_ctx:
            assert isinstance(sub_ctx, RichContext)
            assert sub_ctx.help_config.style_option == "#111111"
            assert sub_ctx.help_config.style_argument == "#333333"

        with subcommand3.make_context("pytest-example", [], parent=ctx) as sub_ctx:
            assert isinstance(sub_ctx, RichContext)
            assert sub_ctx.help_config.style_option == "#444444"


def test_global_config_equal_config_defaults() -> None:
    reload(rc)
    config1 = RichHelpConfiguration()
    config2 = RichHelpConfiguration.load_from_globals(rc)
    for k in {*config1.__dict__.keys(), *config2.__dict__.keys()}:
        if k == "highlighter":
            continue
        v1 = config1.__dict__[k]
        v2 = config2.__dict__[k]
        assert v1 == v2, f"{k}: {v1} != {v2}"


def test_config_from_globals_behavior() -> None:
    original_style_option_value = rc.STYLE_OPTION
    rc.STYLE_OPTION = "new-value"

    @command()
    def cli1() -> None:
        pass

    with cli1.make_context("pytest-example", []) as ctx:
        if TYPE_CHECKING:
            assert isinstance(ctx, RichContext)
        assert ctx.help_config.style_option == "new-value"

    @command()
    @rich_config(help_config=RichHelpConfiguration())
    def cli2() -> None:
        pass

    with cli2.make_context("pytest-example", []) as ctx:
        if TYPE_CHECKING:
            assert isinstance(ctx, RichContext)
        assert ctx.help_config.style_option == original_style_option_value

    @command()
    @rich_config(help_config=RichHelpConfiguration.load_from_globals())
    def cli3() -> None:
        pass

    with cli3.make_context("pytest-example", []) as ctx:
        if TYPE_CHECKING:
            assert isinstance(ctx, RichContext)
        assert ctx.help_config.style_option == "new-value"

    @command()
    @rich_config(help_config={})
    def cli4() -> None:
        pass

    with cli4.make_context("pytest-example", []) as ctx:
        if TYPE_CHECKING:
            assert isinstance(ctx, RichContext)
        assert ctx.help_config.style_option == "new-value"


def test_config_is_serializable_and_invertible() -> None:
    config = RichHelpConfiguration()
    config.apply_theme(force_default=True)

    try:
        serialized_data = json.dumps(asdict(config))
    except TypeError as e:
        pytest.fail(f"RichHelpConfiguration is not serializable: Error raised: {e.__class__.__name__}{e.args}")
    else:
        deserialized_data = json.loads(serialized_data)
        config2 = RichHelpConfiguration(**deserialized_data)

        # Correct types.
        config2.style_options_table_padding = tuple(config2.style_options_table_padding)  # type: ignore[arg-type,assignment]
        config2.style_commands_table_padding = tuple(config2.style_commands_table_padding)  # type: ignore[arg-type,assignment]
        config2.style_options_panel_padding = tuple(config2.style_options_panel_padding)  # type: ignore[arg-type,assignment]
        config2.style_commands_panel_padding = tuple(config2.style_commands_panel_padding)  # type: ignore[arg-type,assignment]
        config2.style_commands_table_column_width_ratio = tuple(config2.style_commands_table_column_width_ratio)  # type: ignore[arg-type,assignment]
        config2.padding_header_text = tuple(config2.padding_header_text)  # type: ignore[arg-type,assignment]
        config2.padding_helptext = tuple(config2.padding_helptext)  # type: ignore[arg-type,assignment]
        config2.padding_errors_panel = tuple(config2.padding_errors_panel)  # type: ignore[arg-type,assignment]
        config2.padding_errors_suggestion = tuple(config2.padding_errors_suggestion)  # type: ignore[arg-type,assignment]
        config2.padding_errors_epilogue = tuple(config2.padding_errors_epilogue)  # type: ignore[arg-type,assignment]

        assert config == config2


def test_multiple_rich_config_passes_is_ok() -> None:
    # It's not advisable that users do any of this,
    # but it shouldn't break anything to do so either.
    c = Console()

    @command()
    @rich_config(help_config={"style_option": "a"})
    @rich_config(help_config={"style_option": "b"})  # should be overwritten
    @rich_config(console=c)
    def cli() -> None:
        pass

    assert cli.context_settings.get("rich_help_config") == {"style_option": "a"}
    assert cli.context_settings.get("rich_console") is c

    @rich_config(help_config={"style_option": "c"})
    @rich_config(help_config={"style_option": "d"})  # should be overwritten
    @rich_config(console=c)
    @command()
    def cli2() -> None:
        pass

    assert cli2.context_settings.get("rich_help_config") == {"style_option": "c"}
    assert cli2.context_settings.get("rich_console") is c


def test_custom_console(cli_runner: CliRunner) -> None:
    # It's not advisable that users do any of this,
    # but it shouldn't break anything to do so either.
    f = io.StringIO()
    c = Console(file=f, width=rc.WIDTH)

    @command()
    @rich_config(console=c)
    def cli() -> None:
        """My CLI help text"""

    cli_runner.invoke(cli, "--help")

    assert f.getvalue() == snapshot(
        """\
                                                                                                    \n\
 Usage: cli [OPTIONS]                                                                               \n\
                                                                                                    \n\
 My CLI help text                                                                                   \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )



from __future__ import annotations

from fnmatch import fnmatch
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from click import Context, Parameter

from rich_click._click_types_cache import Argument, Command, Group
from rich_click.rich_help_configuration import ColumnType, CommandColumnType, OptionColumnType
from rich_click.rich_parameter import RichArgument, RichParameter
from rich_click.utils import CommandGroupDict, OptionGroupDict


if TYPE_CHECKING:
    from rich.box import Box
    from rich.panel import Panel
    from rich.style import StyleType
    from rich.table import Table

    from rich_click.rich_command import RichCommand
    from rich_click.rich_context import RichContext
    from rich_click.rich_help_formatter import RichHelpFormatter


ColT = TypeVar("ColT", bound=ColumnType)
CT = TypeVar("CT", Command, Parameter)
GroupType = TypeVar("GroupType", OptionGroupDict, CommandGroupDict)


class RichPanel(Generic[CT, ColT]):
    """RichPanel base class."""

    panel_class: Optional[Type["Panel"]] = None
    table_class: Optional[Type["Table"]] = None
    _highlight: ClassVar[bool] = False
    _object_attr: ClassVar[str] = NotImplemented

    def __init__(
        self,
        name: str,
        *,
        help: Optional[str] = None,
        help_style: Optional["StyleType"] = None,
        table_styles: Optional[Dict[str, Any]] = None,
        panel_styles: Optional[Dict[str, Any]] = None,
        column_types: Optional[List[ColT]] = None,
        inline_help_in_title: Optional[bool] = None,
        title_style: Optional["StyleType"] = None,
    ) -> None:
        """Initialize a RichPanel."""
        self.name = name
        self.help = help
        self.help_style = help_style
        self.table_styles = table_styles or {}
        self.panel_styles = panel_styles or {}
        self.column_types = column_types
        self.inline_help_in_title = inline_help_in_title
        self.title_style = title_style

    @property
    def objects(self) -> List[str]:
        if self._object_attr is NotImplemented:
            raise NotImplementedError()
        return getattr(self, self._object_attr)  # type: ignore[no-any-return]

    def add_object(self, o: str) -> None:
        if self._object_attr is NotImplemented:
            raise NotImplementedError()
        getattr(self, self._object_attr).append(o)

    def get_box(self, box: Optional[Union[str, "Box"]]) -> Optional["Box"]:
        if box is None:
            return None
        from rich_click.rich_box import get_box

        return get_box(box)

    def to_info_dict(self, ctx: Context) -> Dict[str, Any]:
        if self._object_attr is NotImplemented:
            raise NotImplementedError()
        return {
            "name": self.name,
            "type": self.__class__.__name__,
            "help": self.help,
            self._object_attr: [i.name for i in self.get_objects(ctx.command, ctx)],
        }

    @classmethod
    def list_all_objects(cls, ctx: Context) -> List[Tuple[str, CT]]:
        """List all objects of the command that this panel type works with."""
        raise NotImplementedError()

    def get_objects(self, command: Command, ctx: Context) -> Generator[CT, None, None]:
        """List the proper names assigned to the panel."""
        raise NotImplementedError()

    def _get_base_table(self, **defaults: Any) -> "Table":
        if self.table_class is None:
            from rich.table import Table

            self.table_class = Table

        kw: Dict[str, Any] = {
            "highlight": self._highlight,
            "show_header": False,
        }

        kw.update(defaults)
        kw.update(self.table_styles)
        if "box" in kw and kw["box"] is not None:
            kw["box"] = self.get_box(kw.pop("box", None))
        return self.table_class(**kw)

    def get_table(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Table":
        raise NotImplementedError()

    def _get_base_panel(self, table: "Table", **defaults: Any) -> "Panel":

        if self.panel_class is None:
            from rich_click.rich_help_rendering import RichClickRichPanel

            self.panel_class = RichClickRichPanel
        kw = defaults
        kw.update(self.panel_styles)
        if "box" in kw:
            if kw["box"] is None:
                kw.pop("box")
                kw["box"] = self.get_box("SIMPLE")
            else:
                kw["box"] = self.get_box(kw.pop("box", None))

        return self.panel_class(table, **kw)

    def render(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Panel":
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"


class RichOptionPanel(RichPanel[Parameter, OptionColumnType]):
    """Panel for parameters."""

    _highlight: ClassVar[bool] = True
    _object_attr: ClassVar[str] = "options"

    def __init__(
        self,
        name: str,
        options: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a RichOptionPanel."""
        super().__init__(name, **kwargs)
        self.options = options or []

    @classmethod
    def list_all_objects(cls, ctx: Context) -> List[Tuple[str, Parameter]]:
        return [
            (i.opts[0] if getattr(i, "flag_value", None) and i.opts else i.name, i)  # type: ignore[misc]
            for i in ctx.command.get_params(ctx)
        ]

    def get_objects(self, command: Command, ctx: Context) -> Generator[Parameter, None, None]:
        """List the objects assigned to the panel."""
        for opt in self.options:
            # Get the param
            for param in command.get_params(ctx):
                if any([opt in [*param.opts, param.name]]):
                    break
            # Skip if option is not listed in this group
            else:
                continue
            yield param

    def get_table(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Table":
        t_styles = {
            "show_lines": formatter.config.style_options_table_show_lines,
            "leading": formatter.config.style_options_table_leading,
            "box": formatter.config.style_options_table_box,
            "border_style": formatter.config.style_options_table_border_style,
            "row_styles": formatter.config.style_options_table_row_styles,
            "pad_edge": formatter.config.style_options_table_pad_edge,
            "padding": formatter.config.style_options_table_padding,
            "expand": formatter.config.style_options_table_expand,
        }
        table = self._get_base_table(**t_styles)
        rows = []
        for param in self.get_objects(command, ctx):
            from rich_click.rich_help_rendering import get_parameter_rich_table_row

            cols = (
                param.get_rich_table_row(ctx, formatter, self)
                if isinstance(param, RichParameter)
                else get_parameter_rich_table_row(param, ctx, formatter, self)  # type: ignore[arg-type]
            )

            rows.append(cols)

        headers = [i.replace("_", " ").title() for i in formatter.config.options_table_column_types]

        filtered = [(h, c) for h, c in zip(headers, zip(*rows)) if any(cell for cell in c)]
        headers, rows = zip(*filtered) if filtered else ([], [])
        rows = list(map(list, zip(*rows))) if rows else []

        for row in rows:
            table.add_row(*row)

        for col, header in zip(table.columns, headers):
            col.header = header

        return table

    def render(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Panel":
        from rich.text import Text

        from rich_click.rich_help_rendering import RichClickRichPanel

        inner: Any = self.get_table(command, ctx, formatter)

        p_styles: Dict[str, Any] = {
            "border_style": formatter.config.style_options_panel_border,
            "title_align": formatter.config.align_options_panel,
            "box": formatter.config.style_options_panel_box,
            "padding": formatter.config.style_options_panel_padding,
            "style": formatter.config.style_options_panel_style,
        }
        if self.panel_class is None or issubclass(self.panel_class, RichClickRichPanel):
            p_styles["title_padding"] = formatter.config.panel_title_padding

        if self.title_style is None:
            title_style = formatter.config.style_options_panel_title_style
        else:
            title_style = self.title_style

        title = formatter.config.panel_title_string.format(self.name)

        if self.help:
            if self.help_style is None:
                help_style = formatter.config.style_options_panel_help_style
            else:
                help_style = self.help_style
            if self.inline_help_in_title is None:
                inline_help_in_title = formatter.config.panel_inline_help_in_title
            else:
                inline_help_in_title = self.inline_help_in_title

            if inline_help_in_title:

                p_styles["title"] = Text("", overflow="ellipsis").join(
                    [
                        Text(title, style=title_style),
                        Text(formatter.config.panel_inline_help_delimiter),
                        Text(self.help, style=help_style),
                    ]
                )
            else:
                p_styles["title"] = Text(title, style=title_style)
                from rich.containers import Renderables

                inner = Renderables([formatter.rich_text(self.help, style=help_style), inner])
        else:
            p_styles["title"] = Text(title, style=title_style)
        panel = self._get_base_panel(inner, **p_styles)
        return panel


class RichCommandPanel(RichPanel[Command, CommandColumnType]):
    """Panel for parameters."""

    _object_attr: ClassVar[str] = "commands"

    def __init__(
        self,
        name: str,
        commands: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a RichCommandPanel."""
        super().__init__(name, **kwargs)
        self.commands = commands or []

    @classmethod
    def list_all_objects(cls, ctx: Context) -> List[Tuple[str, Command]]:
        if not isinstance(ctx.command, Group):
            return []
        commands = []
        for cmd_name in ctx.command.list_commands(ctx):
            cmd = ctx.command.get_command(ctx, cmd_name)
            if cmd is not None:
                commands.append((cmd_name, cmd))
        return commands

    def get_objects(self, command: Command, ctx: Context) -> Generator[Command, None, None]:
        """List the objects assigned to the panel."""
        if not isinstance(command, Group):
            return

        commands_list = command.list_commands(ctx)
        callback_names = {c.callback.__name__: c for c in command.commands.values() if c.callback is not None}

        for cmd_name in self.commands:

            if cmd_name in commands_list:
                yield command.get_command(ctx, cmd_name)  # type: ignore[misc]
            elif cmd_name in callback_names:
                yield callback_names[cmd_name]
            else:
                continue

    def get_table(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Table":
        t_styles = {
            "show_lines": formatter.config.style_commands_table_show_lines,
            "leading": formatter.config.style_commands_table_leading,
            "box": formatter.config.style_commands_table_box,
            "border_style": formatter.config.style_commands_table_border_style,
            "row_styles": formatter.config.style_commands_table_row_styles,
            "pad_edge": formatter.config.style_commands_table_pad_edge,
            "padding": formatter.config.style_commands_table_padding,
            "expand": formatter.config.style_commands_table_expand,
        }
        table = self._get_base_table(**t_styles)

        # Define formatting in first column, as commands don't match highlighter regex
        # and set column ratio for first and second column, if a ratio has been set
        if formatter.config.style_commands_table_column_width_ratio is None:
            table_column_width_ratio: Union[Tuple[None, None], Tuple[int, int]] = (None, None)
        else:
            table_column_width_ratio = formatter.config.style_commands_table_column_width_ratio

        # TODO
        # columns = self.columns or formatter.config.commands_table_column_types

        table.add_column(style=formatter.config.style_command, no_wrap=True, ratio=table_column_width_ratio[0])
        table.add_column(
            no_wrap=False,
            ratio=table_column_width_ratio[1],
        )

        if not isinstance(command, Group):
            return table

        rows = []

        for cmd in self.get_objects(command, ctx):

            from rich_click.rich_command import RichCommand
            from rich_click.rich_help_rendering import get_command_rich_table_row

            cols = (
                cmd.get_rich_table_row(ctx, formatter, self)
                if isinstance(cmd, RichCommand)
                else get_command_rich_table_row(cmd, ctx, formatter, self)
            )

            rows.append(cols)

        headers = [i.replace("_", " ").title() for i in formatter.config.commands_table_column_types]

        filtered = [(h, c) for h, c in zip(headers, zip(*rows)) if any(cell for cell in c)]
        headers, rows = zip(*filtered) if filtered else ([], [])
        rows = list(map(list, zip(*rows))) if rows else []

        for row in rows:
            table.add_row(*row)

        for col, header in zip(table.columns, headers):
            col.header = header

        return table

    def render(
        self,
        command: "RichCommand",
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
    ) -> "Panel":
        from rich.text import Text

        from rich_click.rich_help_rendering import RichClickRichPanel

        inner: Any = self.get_table(command, ctx, formatter)

        p_styles: Dict[str, Any] = {
            "border_style": formatter.config.style_commands_panel_border,
            "title_align": formatter.config.align_commands_panel,
            "box": formatter.config.style_commands_panel_box,
            "padding": formatter.config.style_commands_panel_padding,
            "style": formatter.config.style_commands_panel_style,
        }
        if self.panel_class is None or issubclass(self.panel_class, RichClickRichPanel):
            p_styles["title_padding"] = formatter.config.panel_title_padding

        if self.title_style is None:
            title_style = formatter.config.style_commands_panel_title_style
        else:
            title_style = self.title_style

        title = formatter.config.panel_title_string.format(self.name)

        if self.help:
            if self.help_style is None:
                help_style = formatter.config.style_commands_panel_help_style
            else:
                help_style = self.help_style
            if self.inline_help_in_title is None:
                inline_help_in_title = formatter.config.panel_inline_help_in_title
            else:
                inline_help_in_title = self.inline_help_in_title

            if inline_help_in_title:

                p_styles["title"] = Text("", overflow="ellipsis").join(
                    [
                        Text(title, style=title_style),
                        Text(formatter.config.panel_inline_help_delimiter),
                        Text(self.help, style=help_style),
                    ]
                )
            else:
                p_styles["title"] = Text(title, style=title_style)
                from rich.containers import Renderables

                inner = Renderables([formatter.rich_text(self.help, style=help_style), inner])
        else:
            p_styles["title"] = Text(title, style=title_style)

        panel = self._get_base_panel(inner, **p_styles)
        return panel


# Using config to define panels is silently deprecated.
# We do not intend on removing this for a very long time, possibly ever.


def _resolve_panels_from_config(
    ctx: "RichContext",
    formatter: "RichHelpFormatter",
    groups: Dict[str, List[GroupType]],
    panel_cls: Type[RichPanel[CT, ColT]],
) -> List[RichPanel[CT, ColT]]:
    """Logic for resolving the groups."""
    # Step 1: get valid name(s) for the command currently being executed
    assert panel_cls._object_attr is not NotImplemented, "RichPanel must have a defined _object_attr"

    cmd_name = ctx.command.name
    _ctx: "RichContext" = ctx
    while _ctx.parent is not None:
        _ctx = _ctx.parent  # type: ignore[assignment]
        cmd_name = f"{_ctx.command.name} {cmd_name}"
    # 'command_path' is sometimes the file name, e.g. hello.py.
    # We also want to make sure that the actual command name is supported as well.
    if cmd_name != ctx.command_path:
        paths = [cmd_name, ctx.command_path]
    else:
        paths = [cmd_name]
    # Also handle 'python -m foo' when the user specifies a key of 'foo':
    if ctx.command_path.startswith("python -m "):
        extra = ctx.command_path.replace("python -m ", "", 1)
        paths.append(extra)
    final_groups_list: List[GroupType] = []

    # Step 2: Match currently executing command to keys
    # Assign wildcards, but make sure we do not overwrite anything already defined.
    for mtch in reversed(sorted([_ for _ in groups if any(fnmatch(_path, _) for _path in paths)])):
        wildcard_option_groups = groups[mtch]
        for grp in wildcard_option_groups:
            grp = grp.copy()
            opts: List[str] = grp.get(panel_cls._object_attr, [])  # type: ignore[assignment]
            traversed = []
            for opt in grp.get(panel_cls._object_attr, []):  # type: ignore[attr-defined]

                if grp.get("deduplicate", True) and opt in [
                    _opt
                    for _grp in final_groups_list
                    for _opt in _grp.get(panel_cls._object_attr, [])  # type: ignore[attr-defined]
                ]:
                    opts.remove(opt)
                traversed.append(opt)
            grp[panel_cls._object_attr] = opts  # type: ignore[literal-required]
            grp.pop("deduplicate", None)
            final_groups_list.append(grp)

    return [panel_cls(**grp) for grp in final_groups_list]  # type: ignore[misc,arg-type]


def construct_panels(
    command: "RichCommand",
    ctx: "RichContext",
    formatter: "RichHelpFormatter",
) -> List[RichPanel[Any, Any]]:
    """Construct panels from the command as well as from the old groups config."""
    _show_arguments = formatter.config.show_arguments

    # If only an options or a commands panel is defined,
    # then we respect intra-type sort order but not inter-type sort order.
    defined_commands = False
    defined_options = False
    # Handling ordering is tricky with the "groups" feature.
    # it's safest to flag and treat specially.
    using_groups_feat = False

    # Start with list of panels already defined.
    defined_panels: Dict[Tuple[str, str], RichPanel[Any, Any]] = {}

    for p in command.panels:
        defined_panels[(p._object_attr, p.name)] = p
        if p._object_attr == "options":
            defined_options = True
        elif p._object_attr == "commands":
            defined_commands = True

    if formatter.config.option_groups:
        option_groups_from_config = _resolve_panels_from_config(
            ctx, formatter, formatter.config.option_groups, formatter.option_panel_class
        )
        if option_groups_from_config:
            defined_panels.update({(p._object_attr, p.name): p for p in option_groups_from_config})
            using_groups_feat = True

    if isinstance(command, Group):
        if formatter.config.command_groups:
            command_groups_from_config = _resolve_panels_from_config(
                ctx, formatter, formatter.config.command_groups, formatter.command_panel_class
            )
            if command_groups_from_config:
                defined_panels.update({(p._object_attr, p.name): p for p in command_groups_from_config})
                using_groups_feat = True

    # Separate out default panels because we need to sort them properly later.
    # We will reversed() through this so order is flipped.
    # Also-- we have to decouple name from obj because commands can have different names
    # than their mappings to a Group.
    new_panels: Dict[Tuple[str, str], List[str]] = {}
    pre_default_panels: Dict[Tuple[str, str], List[str]] = {}
    post_default_panels: Dict[Tuple[str, str], List[str]]
    if isinstance(command, Group):
        if formatter.config.commands_before_options:
            if defined_commands != defined_options or using_groups_feat:
                pre_default_panels = {
                    ("commands", formatter.config.commands_panel_title): [],
                }
                post_default_panels = {
                    ("options", formatter.config.arguments_panel_title): [],
                    ("options", formatter.config.options_panel_title): [],
                }
            else:
                post_default_panels = {
                    ("commands", formatter.config.commands_panel_title): [],
                    ("options", formatter.config.arguments_panel_title): [],
                    ("options", formatter.config.options_panel_title): [],
                }
        else:
            if defined_commands != defined_options or using_groups_feat:
                pre_default_panels = {
                    ("options", formatter.config.arguments_panel_title): [],
                    ("options", formatter.config.options_panel_title): [],
                }
                post_default_panels = {
                    ("commands", formatter.config.commands_panel_title): [],
                }
            else:
                post_default_panels = {
                    ("options", formatter.config.arguments_panel_title): [],
                    ("options", formatter.config.options_panel_title): [],
                    ("commands", formatter.config.commands_panel_title): [],
                }
    else:
        post_default_panels = {
            ("options", formatter.config.arguments_panel_title): [],
            ("options", formatter.config.options_panel_title): [],
        }

    if ("commands", formatter.config.commands_panel_title) in defined_panels:
        pre_default_panels.pop(("commands", formatter.config.commands_panel_title), None)
        post_default_panels.pop(("commands", formatter.config.commands_panel_title), None)

    if ("options", formatter.config.options_panel_title) in defined_panels:
        pre_default_panels.pop(("options", formatter.config.options_panel_title), None)
        post_default_panels.pop(("options", formatter.config.options_panel_title), None)

    if ("options", formatter.config.arguments_panel_title) in defined_panels:
        pre_default_panels.pop(("options", formatter.config.arguments_panel_title), None)
        post_default_panels.pop(("options", formatter.config.arguments_panel_title), None)

    # Go through objects to see whether they are assigned.
    # Need to do tuples because commands and options can have same name.
    assigned_objects: Dict[Tuple[str, str], Set[str]] = {}
    for p in defined_panels.values():
        for o in p.objects:
            assigned_objects.setdefault((p._object_attr, o), set())
            assigned_objects[(p._object_attr, o)].add(p.name)

    if ("options", formatter.config.arguments_panel_title) in defined_panels:
        _show_arguments = True

    objs: List[Tuple[str, str, Union[Parameter, Command]]] = [
        ("options", name, o) for name, o in formatter.option_panel_class.list_all_objects(ctx)
    ]
    if isinstance(command, Group):
        objs.extend([("commands", name, o) for name, o in formatter.command_panel_class.list_all_objects(ctx)])

    from rich_click.rich_command import RichGroup

    # Here we are interested in:
    # 1. assigning objs based on panel=...
    # 2. getting unassigned objs
    for typ, name, obj in objs:
        if TYPE_CHECKING:
            assert isinstance(obj.name, str)
        if getattr(obj, "hidden", False):
            continue
        names = {name, obj.name}
        if isinstance(obj, Parameter):
            names.update(obj.opts)
        elif isinstance(obj, Command) and obj.callback is not None:
            names.add(obj.callback.__name__)
        assigned_to = set()
        for n in names:
            _assigned_panels = assigned_objects.get((typ, n), set())
            for ap in _assigned_panels:
                assigned_to.add((typ, ap))
        assigned = bool(assigned_to)
        inferred = False
        panel_list: List[str] = []
        if hasattr(obj, "panel"):
            if isinstance(obj.panel, str):
                panel_list = [obj.panel]
            elif obj.panel is None:
                if assigned:
                    continue
            else:
                panel_list = obj.panel
        elif typ == "commands" and isinstance(command, RichGroup):
            _p = command._panel_command_mapping.get(name)
            if _p:
                panel_list.extend(_p)
        if not panel_list:
            inferred = True
            if typ == "options":
                if not formatter.config.group_arguments_options and isinstance(obj, Argument):
                    if _show_arguments is not False:
                        panel_list = [formatter.config.arguments_panel_title]
                    else:
                        panel_list = []
                else:
                    panel_list = [formatter.config.options_panel_title]
            elif typ == "commands":
                panel_list = [formatter.config.commands_panel_title]
            else:
                panel_list = []
        for panel_name in panel_list:
            # Ensure we don't reassign if already assigned.
            if (typ, panel_name) not in assigned_to:
                if (typ, panel_name) in defined_panels:
                    defined_panels[(typ, panel_name)].add_object(name)
                elif assigned:
                    pass
                elif (typ, panel_name) in pre_default_panels:
                    pre_default_panels[(typ, panel_name)].append(name)
                elif (typ, panel_name) in post_default_panels:
                    post_default_panels[(typ, panel_name)].append(name)
                else:
                    new_panels.setdefault((typ, panel_name), []).append(name)
            if (
                _show_arguments is None
                and panel_name == formatter.config.arguments_panel_title
                and isinstance(obj, RichArgument)
                and (not inferred or obj.help is not None)
            ):
                _show_arguments = True

    if not _show_arguments:
        pre_default_panels.pop(("options", formatter.config.arguments_panel_title), None)
        post_default_panels.pop(("options", formatter.config.arguments_panel_title), None)

    final_panels: List[RichPanel[Any, Any]] = []

    all_panel_mappings: List[Union[Dict[Tuple[str, str], List[str]], Dict[Tuple[str, str], RichPanel[Any, Any]]]] = [
        pre_default_panels,
        defined_panels,
        new_panels,
        post_default_panels,
    ]
    for d in all_panel_mappings:
        for (typ, panel_name), obj_list in d.items():
            cls: Type[RichPanel[Any, Any]]
            if typ == "options":
                cls = formatter.option_panel_class
            elif typ == "commands":
                cls = formatter.command_panel_class
            else:
                continue
            panel: RichPanel[Any, Any]
            if isinstance(obj_list, RichPanel):
                panel = obj_list
            elif (typ, panel_name) not in defined_panels:
                panel = cls(panel_name)
                setattr(panel, panel._object_attr, [i for i in obj_list])
            else:
                panel = defined_panels[(typ, panel_name)]
                for _obj in obj_list:
                    panel.add_object(_obj)
            final_panels.append(panel)

    return final_panels



from __future__ import annotations

import io
import sys
from contextlib import contextmanager
from functools import cached_property
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import click

from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_panel import RichCommandPanel, RichOptionPanel, RichPanel


if TYPE_CHECKING:  # pragma: no cover
    from rich.console import Console
    from rich.highlighter import Highlighter
    from rich.markdown import Markdown
    from rich.style import StyleType
    from rich.text import Text


RP = TypeVar("RP", bound=RichPanel[Any, Any])


class _EncodedStringIO(io.StringIO):
    """StringIO with encoding attributes for Rich Console detection."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._encoding = getattr(sys.stdout, "encoding", None) or "utf-8"
        self._errors = getattr(sys.stdout, "errors", None) or "replace"

    @property
    def encoding(self) -> str:  # type: ignore[override]
        return self._encoding

    @property
    def errors(self) -> str:  # type: ignore[override]
        return self._errors


def create_console(
    config: RichHelpConfiguration,
    file: Optional[IO[str]] = None,
    width: Optional[int] = None,
    max_width: Optional[int] = None,
) -> "Console":
    """
    Create a Rich Console configured from Rich Help Configuration.

    Args:
    ----
        config: Rich Help Configuration instance
        file: Optional IO stream to write Rich Console output
            Defaults to None.
        width: Width of the Console; overrides config.width if set.
        max_width: Max width of the Console; overrides config.max_width if set.

    """
    from rich.console import Console
    from rich.theme import Theme

    console = Console(
        theme=Theme(
            {
                "option": config.style_option,
                "command": config.style_command,
                "argument": config.style_argument,
                "switch": config.style_switch,
                "metavar": config.style_metavar,
                "metavar_sep": config.style_metavar_separator,
                "usage": config.style_usage,
                "deprecated": config.style_deprecated,
            }
        ),
        color_system=config.color_system,
        force_terminal=config.force_terminal,
        width=width if width is not None else config.width,
        record=True if file is None else False,
        legacy_windows=config.legacy_windows,
    )
    # Defaults for console.color_system change when file is in __init__.
    # Workaround: set file after __init__.
    console.file = file or _EncodedStringIO()
    max_width = max_width if max_width is not None else config.max_width
    if isinstance(max_width, int):
        console.width = min(max_width, console.size.width)
    return console


class RichHelpFormatter(click.HelpFormatter):
    """
    Rich Help Formatter.

    This class is a container for the help configuration and Rich Console that
    are used internally by the help and error printing methods.
    """

    console: "Console"
    """Rich Console created from the help configuration.

    This console is meant only for use with the formatter and should
    not be created directly
    """
    export_console_as: Literal[None, "html", "svg", "text"] = None

    option_panel_class: Type[RichOptionPanel] = RichOptionPanel
    command_panel_class: Type[RichCommandPanel] = RichCommandPanel

    def __init__(
        self,
        indent_increment: int = 2,
        width: Optional[int] = None,
        max_width: Optional[int] = None,
        *args: Any,
        console: Optional["Console"] = None,
        config: Optional[RichHelpConfiguration] = None,
        export_console_as: Literal[None, "html", "svg", "text"] = None,
        export_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Create Rich Help Formatter.

        Args:
        ----
            indent_increment: Passed to click.HelpFormatter.
            width: Passed to click.HelpFormatter. Overrides config.width if not None.
            max_width: Passed to click.HelpFormatter. Overrides config.max_width if not None.
            *args: Args passed to click.HelpFormatter.
            console: Use an external console.
            config: RichHelpConfiguration. If None, then build config from globals.
            file: Stream to output to in the Rich Console. If None, use stdout.
            export_console_as: How output is rendered by getvalue(). Default of None renders output normally.
            export_kwargs: Any kwargs passed to the export method of the Console in getvalue().
            **kwargs: Kwargs passed to click.HelpFormatter.

        """
        if config is not None:
            self.config = config
            # Rich config overrides width and max width if set.
        else:
            self.config = RichHelpConfiguration.load_from_globals()

        self.config.apply_theme(force_default=True)

        file = kwargs.pop("file", None)
        if file is not None:
            import warnings

            warnings.warn(
                "The file kwarg to `RichHelpFormatter()` is deprecated" " and will be removed in a future release.",
                DeprecationWarning,
                stacklevel=2,
            )

        if console:
            self.console = console
        else:
            self.console = create_console(self.config, file=file, width=width, max_width=max_width)

        width = self.console.width

        self.export_console_as = export_console_as
        self.export_kwargs = export_kwargs or {}

        super().__init__(indent_increment, width, max_width, *args, **kwargs)

    @property
    def width(self) -> int:
        return self.console.width

    @width.setter
    def width(self, v: int) -> None:
        self.console.width = v

    @cached_property
    def highlighter(self) -> "Highlighter":
        if self.config.highlighter is not None:
            return self.config.highlighter
        else:
            from rich.highlighter import RegexHighlighter

            class HighlighterClass(RegexHighlighter):
                highlights = self.config.highlighter_patterns

            return HighlighterClass()

    def write(self, *objects: Any, **kwargs: Any) -> None:
        self.console.print(*objects, **kwargs)

    def write_usage(self, prog: str, args: str = "", prefix: Optional[str] = None) -> None:
        from rich_click.rich_help_rendering import get_rich_usage

        get_rich_usage(formatter=self, prog=prog, args=args, prefix=prefix)

    def write_error(self, e: click.ClickException) -> None:
        from rich_click.rich_help_rendering import rich_format_error

        rich_format_error(self=e, formatter=self)

    def write_abort(self) -> None:
        """Print richly formatted abort error."""
        self.console.print(self.config.aborted_text, style=self.config.style_aborted)

    def rich_text(
        self,
        text: Union[str, "Text", "Markdown"],
        style: "StyleType" = "",
    ) -> Union["Text", "Markdown"]:
        """
        Take a string, remove indentations, and return styled text.
        By default, return the text as a Rich Text with the request style.

        This method uses the config options prefixed `text_` to influence how
        rendering works.

        Args:
        ----
            text (str): Text to style.
            style (StyleType): Rich style to apply.

        Returns:
        -------
            MarkdownElement or Text: Styled text object

        """
        import inspect

        from rich.jupyter import JupyterMixin
        from rich.text import Text

        if isinstance(text, JupyterMixin):
            return text

        # Remove indentations from input text
        text = inspect.cleandoc(text)

        kw: Dict[str, Any]
        if self.config.text_markup != "rich":
            kw = {"style": style}
            if self.config.text_emojis:
                from rich.emoji import Emoji

                text = Emoji.replace(text)
        else:
            kw = {"style": style, "emoji": self.config.text_emojis}

        kw.update(self.config.text_kwargs or {})

        if self.config.text_markup == "markdown":
            # Lazy load Markdown because it slows down rendering
            from rich.markdown import Markdown

            return Markdown(text, **kw)
        elif self.config.text_markup == "rich":
            return self.highlighter(Text.from_markup(text, **kw))
        elif self.config.text_markup == "ansi":
            return self.highlighter(Text.from_ansi(text, **kw))
        else:
            return self.highlighter(Text(text, **kw))

    def getvalue(self) -> str:
        if not self.export_console_as:
            from rich.console import COLOR_SYSTEMS
            from rich.segment import Segment

            if self.console.legacy_windows:
                # Handle legacy windows
                import colorama  # type: ignore[import-untyped]

                colorama.init()

            if self.console.no_color:
                res = "".join(
                    (
                        style.render(
                            text,
                            color_system=(
                                COLOR_SYSTEMS.get(self.console.color_system) if self.console.color_system else None
                            ),
                        )
                        if style
                        else text
                    )
                    for text, style, _ in Segment.remove_color(self.console._record_buffer)
                )
            else:
                res = "".join(
                    (
                        style.render(
                            text,
                            color_system=(
                                COLOR_SYSTEMS.get(self.console.color_system) if self.console.color_system else None
                            ),
                        )
                        if style
                        else text
                    )
                    for text, style, _ in self.console._record_buffer
                )
            return res
        elif self.console.record:
            kw = self.export_kwargs.copy()
            kw.setdefault("clear", False)
            if self.export_console_as == "text" or self.export_console_as is None:
                res = self.console.export_text(**kw)
            elif self.export_console_as == "html":
                kw.setdefault("inline_styles", True)
                res = self.console.export_html(**kw)
            elif self.export_console_as == "svg":
                kw.setdefault("title", " ".join(sys.argv))
                res = self.console.export_svg(**kw)
            else:
                raise ValueError(
                    "Invalid value for `export_console_as`." " Must be one of 'text', 'html', 'svg', or None."
                )
            return res
        else:
            return super().getvalue()

    def indent(self) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    def dedent(self) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    def write_heading(self, heading: str) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    def write_paragraph(self) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    def write_text(self, text: str) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    def write_dl(
        self,
        rows: Sequence[Tuple[str, str]],
        col_max: int = 30,
        col_spacing: int = 2,
    ) -> None:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)

    @contextmanager
    def section(self, name: str) -> Iterator[None]:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)
        yield

    @contextmanager
    def indentation(self) -> Iterator[None]:
        import warnings

        warnings.warn("Not implemented.", RuntimeWarning, stacklevel=2)
        yield



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple, Union

from rich_click.rich_help_configuration import (
    FROM_THEME,
    CommandColumnType,
    CommandHelpSectionType,
    OptionColumnType,
    OptionHelpSectionType,
    force_terminal_default,
    terminal_width_default,
)
from rich_click.utils import CommandGroupDict, OptionGroupDict, notset


if TYPE_CHECKING:  # pragma: no cover
    from rich.align import AlignMethod
    from rich.box import Box
    from rich.padding import PaddingDimensions
    from rich.style import StyleType
    from rich.text import Text

    from rich_click.rich_click_theme import RichClickTheme

#!STARTCONFIG

# Default styles
THEME: Optional[Union[str, RichClickTheme]] = None
ENABLE_THEME_ENV_VAR: bool = True

STYLE_OPTION: "StyleType" = FROM_THEME
STYLE_OPTION_NEGATIVE: Optional["StyleType"] = FROM_THEME
STYLE_ARGUMENT: "StyleType" = FROM_THEME
STYLE_COMMAND: "StyleType" = FROM_THEME
STYLE_COMMAND_ALIASES: "StyleType" = FROM_THEME
STYLE_SWITCH: "StyleType" = FROM_THEME
STYLE_SWITCH_NEGATVE: Optional["StyleType"] = FROM_THEME
STYLE_METAVAR: "StyleType" = FROM_THEME
STYLE_METAVAR_APPEND: "StyleType" = FROM_THEME
STYLE_METAVAR_SEPARATOR: "StyleType" = FROM_THEME
STYLE_RANGE_APPEND: "StyleType" = FROM_THEME
STYLE_USAGE: "StyleType" = FROM_THEME
STYLE_USAGE_COMMAND: "StyleType" = FROM_THEME
STYLE_USAGE_SEPARATOR: "StyleType" = FROM_THEME
STYLE_DEPRECATED: "StyleType" = FROM_THEME
STYLE_HELPTEXT_FIRST_LINE: "StyleType" = FROM_THEME
STYLE_HELPTEXT: "StyleType" = FROM_THEME
STYLE_HELPTEXT_ALIASES: Optional["StyleType"] = None
STYLE_OPTION_HELP: "StyleType" = FROM_THEME
STYLE_COMMAND_HELP: "StyleType" = FROM_THEME
STYLE_OPTION_DEFAULT: "StyleType" = FROM_THEME
STYLE_OPTION_ENVVAR: "StyleType" = FROM_THEME
STYLE_REQUIRED_SHORT: "StyleType" = FROM_THEME
STYLE_REQUIRED_LONG: "StyleType" = FROM_THEME
STYLE_OPTIONS_PANEL_BORDER: "StyleType" = FROM_THEME
STYLE_OPTIONS_PANEL_BOX: Optional[Union[str, "Box"]] = FROM_THEME
STYLE_OPTIONS_PANEL_HELP_STYLE: "StyleType" = FROM_THEME
STYLE_OPTIONS_PANEL_TITLE_STYLE: "StyleType" = FROM_THEME
STYLE_OPTIONS_PANEL_PADDING: "PaddingDimensions" = FROM_THEME
STYLE_OPTIONS_PANEL_STYLE: "StyleType" = FROM_THEME
ALIGN_OPTIONS_PANEL: "AlignMethod" = "left"
STYLE_OPTIONS_TABLE_SHOW_LINES: bool = False
STYLE_OPTIONS_TABLE_LEADING: int = 0
STYLE_OPTIONS_TABLE_PAD_EDGE: bool = False
STYLE_OPTIONS_TABLE_PADDING: "PaddingDimensions" = (0, 1)
STYLE_OPTIONS_TABLE_EXPAND: bool = FROM_THEME
STYLE_OPTIONS_TABLE_BOX: Optional[Union[str, "Box"]] = FROM_THEME
STYLE_OPTIONS_TABLE_ROW_STYLES: Optional[List["StyleType"]] = None
STYLE_OPTIONS_TABLE_BORDER_STYLE: Optional["StyleType"] = FROM_THEME
STYLE_COMMANDS_PANEL_BORDER: "StyleType" = FROM_THEME
PANEL_INLINE_HELP_IN_TITLE: bool = FROM_THEME
PANEL_INLINE_HELP_DELIMITER: str = FROM_THEME
STYLE_COMMANDS_PANEL_BOX: Optional[Union[str, "Box"]] = FROM_THEME
STYLE_COMMANDS_PANEL_HELP_STYLE: "StyleType" = FROM_THEME
STYLE_COMMANDS_PANEL_TITLE_STYLE: "StyleType" = FROM_THEME
STYLE_COMMANDS_PANEL_PADDING: "PaddingDimensions" = FROM_THEME
STYLE_COMMANDS_PANEL_STYLE: "StyleType" = FROM_THEME
ALIGN_COMMANDS_PANEL: "AlignMethod" = "left"
STYLE_COMMANDS_TABLE_SHOW_LINES: bool = False
STYLE_COMMANDS_TABLE_LEADING: int = 0
STYLE_COMMANDS_TABLE_PAD_EDGE: bool = False
STYLE_COMMANDS_TABLE_PADDING: "PaddingDimensions" = (0, 1)
STYLE_COMMANDS_TABLE_EXPAND: bool = FROM_THEME
STYLE_COMMANDS_TABLE_BOX: Optional[Union[str, "Box"]] = FROM_THEME
STYLE_COMMANDS_TABLE_ROW_STYLES: Optional[List["StyleType"]] = None
STYLE_COMMANDS_TABLE_BORDER_STYLE: Optional["StyleType"] = FROM_THEME
STYLE_COMMANDS_TABLE_COLUMN_WIDTH_RATIO: Optional[Union[Tuple[None, None], Tuple[int, int]]] = (None, None)
STYLE_ERRORS_PANEL_BORDER: "StyleType" = FROM_THEME
STYLE_ERRORS_PANEL_BOX: Optional[Union[str, "Box"]] = FROM_THEME
ALIGN_ERRORS_PANEL: "AlignMethod" = "left"
STYLE_ERRORS_SUGGESTION: Optional["StyleType"] = None
STYLE_ERRORS_SUGGESTION_COMMAND: Optional["StyleType"] = None
STYLE_PADDING_ERRORS: "StyleType" = FROM_THEME
STYLE_ABORTED: "StyleType" = "red"
STYLE_PADDING_USAGE: "StyleType" = FROM_THEME
STYLE_PADDING_HELPTEXT: "StyleType" = FROM_THEME
STYLE_PADDING_EPILOG: "StyleType" = FROM_THEME
STYLE_HEADER_TEXT: "StyleType" = FROM_THEME
STYLE_EPILOG_TEXT: "StyleType" = FROM_THEME
STYLE_FOOTER_TEXT: "StyleType" = FROM_THEME

PANEL_TITLE_PADDING: int = FROM_THEME
WIDTH: Optional[int] = terminal_width_default()
MAX_WIDTH: Optional[int] = terminal_width_default()
COLOR_SYSTEM: Optional[Literal["auto", "standard", "256", "truecolor", "windows"]] = (
    "auto"  # Set to None to disable colors
)
FORCE_TERMINAL: Optional[bool] = force_terminal_default()

OPTIONS_TABLE_COLUMN_TYPES: List[OptionColumnType] = FROM_THEME
COMMANDS_TABLE_COLUMN_TYPES: List[CommandColumnType] = FROM_THEME

OPTIONS_TABLE_HELP_SECTIONS: List[OptionHelpSectionType] = FROM_THEME
COMMANDS_TABLE_HELP_SECTIONS: List[CommandHelpSectionType] = FROM_THEME

# Fixed strings
HEADER_TEXT: Optional[Union[str, "Text"]] = None
FOOTER_TEXT: Optional[Union[str, "Text"]] = None
PANEL_TITLE_STRING: str = FROM_THEME
DEPRECATED_STRING: str = FROM_THEME
DEPRECATED_WITH_REASON_STRING: str = FROM_THEME
DEFAULT_STRING: str = FROM_THEME
ENVVAR_STRING: str = FROM_THEME
REQUIRED_SHORT_STRING: str = FROM_THEME
REQUIRED_LONG_STRING: str = FROM_THEME
RANGE_STRING: str = FROM_THEME
APPEND_METAVARS_HELP_STRING: str = FROM_THEME
APPEND_RANGE_HELP_STRING: str = FROM_THEME
HELPTEXT_ALIASES_STRING: str = "Aliases: {}"
ARGUMENTS_PANEL_TITLE: str = "Arguments"
OPTIONS_PANEL_TITLE: str = "Options"
COMMANDS_PANEL_TITLE: str = "Commands"
ERRORS_PANEL_TITLE: str = "Error"
DELIMITER_COMMA: str = FROM_THEME
DELIMITER_SLASH: str = FROM_THEME
ERRORS_SUGGESTION: Optional[Union[str, "Text"]] = None  # Default: Try 'cmd -h' for help. Set to False to disable.
ERRORS_EPILOGUE: Optional[Union[str, "Text"]] = None
ABORTED_TEXT: str = "Aborted."

PADDING_HEADER_TEXT: "PaddingDimensions" = FROM_THEME
PADDING_USAGE: "PaddingDimensions" = FROM_THEME
PADDING_HELPTEXT: "PaddingDimensions" = FROM_THEME
PADDING_HELPTEXT_DEPRECATED: "PaddingDimensions" = 0
PADDING_HELPTEXT_FIRST_LINE: "PaddingDimensions" = 0
PADDING_EPILOG: "PaddingDimensions" = FROM_THEME
PADDING_FOOTER_TEXT: "PaddingDimensions" = FROM_THEME
PADDING_ERRORS_PANEL: "PaddingDimensions" = (0, 0, 1, 0)
PADDING_ERRORS_SUGGESTION: "PaddingDimensions" = (0, 1, 0, 1)
PADDING_ERRORS_EPILOGUE: "PaddingDimensions" = (0, 1, 1, 1)

# Behaviours
SHOW_ARGUMENTS: Optional[bool] = None  # Show positional arguments
SHOW_METAVARS_COLUMN: Optional[bool] = None  # Show a column with the option metavar (eg. INTEGER)
COMMANDS_BEFORE_OPTIONS: bool = False  # If set, the commands panel show above the options panel.
APPEND_METAVARS_HELP: Optional[bool] = None  # Append metavar (eg. [TEXT]) after the help text
GROUP_ARGUMENTS_OPTIONS: bool = False  # Show arguments with options instead of in own panel
OPTION_ENVVAR_FIRST: Optional[Optional[bool]] = None  # Show env vars before option help text instead of avert
TEXT_MARKUP: Literal["ansi", "rich", "markdown", None] = notset
TEXT_KWARGS: Optional[Dict[str, Any]] = None
TEXT_EMOJIS: bool = notset
TEXT_PARAGRAPH_LINEBREAKS: Optional[Literal["\n", "\n\n"]] = None
# If set, parse emoji codes and replace with actual emojis, e.g. :smiley_cat: -> 😺
USE_MARKDOWN: Optional[bool] = None  # Parse help strings as markdown
USE_MARKDOWN_EMOJI: Optional[bool] = None  # Parse emoji codes in markdown :smile:
USE_RICH_MARKUP: Optional[bool] = None  # Parse help strings for rich markup (eg. [red]my text[/])
# Define sorted groups of panels to display subcommands
COMMAND_GROUPS: Dict[str, List[CommandGroupDict]] = {}
# Define sorted groups of panels to display options and arguments
OPTION_GROUPS: Dict[str, List[OptionGroupDict]] = {}
USE_CLICK_SHORT_HELP: bool = False  # Use click's default function to truncate help text
HELPTEXT_SHOW_ALIASES: bool = True

#!ENDCONFIG

_THEME_FROM_CLI: Optional[str] = None


def __getattr__(name: str) -> Any:
    if name == "get_module_help_configuration":
        import warnings

        warnings.warn(
            "get_module_help_configuration() is deprecated. Use RichHelpConfiguration.load_from_globals() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from rich_click.rich_help_configuration import RichHelpConfiguration

        return RichHelpConfiguration.load_from_globals
    if name == "highlighter":
        import warnings

        warnings.warn(
            "`highlighter` config option is deprecated."
            " Please do one of the following instead: either set HIGHLIGHTER_PATTERNS = [...] if you want"
            " to use regex; or for more advanced use cases where you'd like to use a different type"
            " of rich.highlighter.Highlighter, subclass the `RichHelpFormatter` and update its `highlighter`.",
            DeprecationWarning,
            stacklevel=2,
        )

        from rich_click.rich_help_configuration import OptionHighlighter

        globals()["highlighter"] = highlighter = OptionHighlighter()
        return highlighter

    elif name in {
        "_make_rich_rext",
        "_get_help_text",
        "_get_option_help",
        "_make_command_help",
        "get_rich_usage",
        "rich_format_help",
        "rich_format_error",
        "rich_abort_error",
    }:
        import warnings

        warnings.warn(
            f"{name}() is no longer located in the `rich_click` module. It is now in the `rich_help_rendering` module.",
            DeprecationWarning,
            stacklevel=2,
        )
        import rich_click.rich_help_rendering

        return getattr(rich_click.rich_help_rendering, name)
    else:
        raise AttributeError



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Mapping, Optional, Type, Union

import click
from click.globals import get_current_context as click_get_current_context

from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_help_formatter import RichHelpFormatter


if TYPE_CHECKING:  # pragma: no cover
    from types import TracebackType

    from rich.console import Console


class RichContext(click.Context):
    """Click Context class endowed with Rich superpowers."""

    formatter_class: Type[RichHelpFormatter] = RichHelpFormatter
    console: Optional["Console"] = None
    export_console_as: Optional[Literal["html", "svg", "text"]] = None
    errors_in_output_format: bool = False
    help_to_stderr: bool = False

    def __init__(
        self,
        *args: Any,
        rich_console: Optional["Console"] = None,
        rich_help_config: Optional[Union[Mapping[str, Any], RichHelpConfiguration]] = None,
        export_console_as: Optional[Literal["html", "svg", "text"]] = None,
        errors_in_output_format: Optional[bool] = None,
        help_to_stderr: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """
        Create Rich Context instance.

        Args:
        ----
            *args: Args that get passed to click.Context.
            rich_console: Rich Console. Defaults to None.
            rich_help_config: Rich help configuration.  Defaults to None.
            export_console_as: Arg is passed to RichHelpFormatter().
            errors_in_output_format: Arg is passed to RichHelpFormatter().
            help_to_stderr: If set, help is printed to stderr.
            **kwargs: Kwargs that get passed to click.Context.

        """
        super().__init__(*args, **kwargs)
        parent: Optional[RichContext] = kwargs.pop("parent", None)

        if help_to_stderr is None and hasattr(parent, "help_to_stderr"):
            self.help_to_stderr = parent.help_to_stderr  # type: ignore[union-attr]
        else:
            self.help_to_stderr = help_to_stderr or self.help_to_stderr

        if export_console_as is None and hasattr(parent, "export_console_as"):
            self.export_console_as = parent.export_console_as  # type: ignore[union-attr]
        else:
            self.export_console_as = export_console_as or self.export_console_as

        if errors_in_output_format is None and hasattr(parent, "export_console_as"):
            self.errors_in_output_format = parent.errors_in_output_format  # type: ignore[union-attr]
        else:
            self.errors_in_output_format = errors_in_output_format or self.errors_in_output_format

        if rich_console is None and hasattr(parent, "console"):
            rich_console = parent.console  # type: ignore[union-attr]

        if rich_console is not None:
            self.console = rich_console

        if rich_help_config is None:
            if hasattr(parent, "help_config"):
                self.help_config = parent.help_config  # type: ignore[has-type,union-attr]
            else:
                self.help_config = RichHelpConfiguration.load_from_globals()
        elif isinstance(rich_help_config, Mapping):
            if hasattr(parent, "help_config"):
                if TYPE_CHECKING:  # pragma: no cover
                    assert parent is not None
                kw = parent.help_config.__dict__.copy()
                kw.update(rich_help_config)
                self.help_config = RichHelpConfiguration(**kw)
            else:
                self.help_config = RichHelpConfiguration.load_from_globals(**rich_help_config)
        else:
            self.help_config = rich_help_config

    def make_formatter(self, error_mode: bool = False) -> RichHelpFormatter:
        """Create the Rich Help Formatter."""
        formatter = self.formatter_class(
            width=self.terminal_width,
            max_width=self.max_content_width,
            config=self.help_config,
            console=self.console,
            export_console_as=(self.export_console_as if not error_mode or self.errors_in_output_format else None),
        )
        return formatter

    if TYPE_CHECKING:  # pragma: no cover

        def __enter__(self) -> "RichContext":
            return super().__enter__()  # type: ignore[return-value]

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            tb: Optional[TracebackType],
        ) -> None:
            super().__exit__(exc_type, exc_value, tb)


def get_current_context(silent: bool = False) -> Optional[RichContext]:
    """
    Return the current click context.  This can be used as a way to
    access the current context object from anywhere.  This is a more implicit
    alternative to the :func:`pass_context` decorator.  This function is
    primarily useful for helpers such as :func:`echo` which might be
    interested in changing its behavior based on the current context.

    To push the current context, :meth:`Context.scope` can be used.

    .. versionadded:: 5.0

    :param silent: if set to `True` the return value is `None` if no context
                   is available.  The default behavior is to raise a
                   :exc:`RuntimeError`.
    """
    return click_get_current_context(silent=silent)  # type: ignore[return-value]



# Some influence take from stubs in Cloup:
# https://github.com/janluke/cloup
# Copyright (c) 2021, Gianluca Gippetto MIT
from __future__ import annotations

import sys
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

import click
from click.shell_completion import CompletionItem
from click.types import ParamType
from rich.console import Console

from rich_click._internal_types import PanelKwargs, RichContextSettingsDict, RichHelpConfigurationDict, TableKwargs
from rich_click.rich_command import RichCommand, RichGroup
from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import CommandColumnType, OptionColumnType, RichHelpConfiguration
from rich_click.rich_panel import RichOptionPanel, RichPanel

if sys.version_info < (3, 11):
    pass
else:
    pass

if sys.version_info < (3, 10):
    from typing_extensions import Concatenate, ParamSpec
else:
    from typing import Concatenate, ParamSpec

_AnyCallable = Callable[..., Any]

StyleType = str  # Pyright gets upset at rich.style.Style, so just use str
FC = TypeVar("FC", bound=Union[click.Command, _AnyCallable])
P = TypeVar("P", bound=click.Parameter)
C = TypeVar("C", bound=click.Command)
G = TypeVar("G", bound=click.Group)
RP = TypeVar("RP", bound=RichPanel[Any, Any])

ShellCompleteArg = Callable[
    [click.Context, P, str],
    Union[List[CompletionItem], List[str]],
]
ParamDefault = Union[Any, Callable[[], Any]]
ParamCallback = Callable[[click.Context, P, Any], Any]

# variant: no call, directly as decorator for a function.
@overload
def command(name: _AnyCallable) -> RichCommand: ...
@overload
def command(
    name: Optional[str] = ...,
    *,
    cls: Type[C],
    context_settings: RichContextSettingsDict,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    no_args_is_help: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], C]: ...
@overload
def command(
    name: Optional[str] = ...,
    *,
    cls: None,
    context_settings: RichContextSettingsDict,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    no_args_is_help: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], RichCommand]: ...
@overload
def command(
    name: Optional[str] = ...,
    *,
    cls: Type[C],
    context_settings: MutableMapping[str, Any] | None = ...,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    no_args_is_help: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], C]: ...
@overload
def command(
    name: Optional[str] = None,
    *,
    cls: None,
    context_settings: MutableMapping[str, Any] | None = ...,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    no_args_is_help: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
    **attrs: Any,
) -> Callable[[_AnyCallable], RichCommand]: ...

# variant: with positional name and with positional or keyword cls argument:
# @command(namearg, CommandCls, ...) or @command(namearg, cls=CommandCls, ...)
@overload
def command(
    name: Optional[str],
    cls: Type[C],
    **attrs: Any,
) -> Callable[[_AnyCallable], C]: ...

# variant: name omitted, cls _must_ be a keyword argument, @command(cls=CommandCls, ...)
@overload
def command(
    name: None = None,
    *,
    cls: Type[C],
    **attrs: Any,
) -> Callable[[_AnyCallable], C]: ...

# variant: with optional string name, no cls argument provided.
@overload
def command(name: Optional[str] = ..., cls: None = None, **attrs: Any) -> Callable[[_AnyCallable], RichCommand]: ...
def command(
    name: Optional[str] = None, *, cls: Optional[Type[C]] = None, **kwargs: Any
) -> Callable[[_AnyCallable], Union[click.Command, C]]: ...

# variant: no call, directly as decorator for a function.
@overload
def group(name: _AnyCallable) -> RichGroup: ...
@overload
def group(
    name: Optional[str] = None,
    *,
    cls: Type[G],
    commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
    invoke_without_command: bool = ...,
    no_args_is_help: bool | None = ...,
    subcommand_metavar: str | None = ...,
    chain: bool = ...,
    result_callback: Callable[..., Any] | None = ...,
    context_settings: RichContextSettingsDict,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], G]: ...
@overload
def group(
    name: Optional[str] = ...,
    *,
    cls: None,
    commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
    invoke_without_command: bool = ...,
    no_args_is_help: bool | None = ...,
    subcommand_metavar: str | None = ...,
    chain: bool = ...,
    result_callback: Callable[..., Any] | None = ...,
    context_settings: MutableMapping[str, Any] | None = ...,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], RichGroup]: ...
@overload
def group(
    name: Optional[str] = None,
    *,
    cls: Type[G],
    commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
    invoke_without_command: bool = ...,
    no_args_is_help: bool | None = ...,
    subcommand_metavar: str | None = ...,
    chain: bool = ...,
    result_callback: Callable[..., Any] | None = ...,
    context_settings: MutableMapping[str, Any] | None = ...,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], G]: ...
@overload
def group(
    name: Optional[str] = ...,
    *,
    cls: None,
    commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
    invoke_without_command: bool = ...,
    no_args_is_help: bool | None = ...,
    subcommand_metavar: str | None = ...,
    chain: bool = ...,
    result_callback: Callable[..., Any] | None = ...,
    context_settings: RichContextSettingsDict,
    callback: Callable[..., Any] | None = ...,
    params: list[click.Parameter] | None = ...,
    help: str | None = ...,
    epilog: str | None = ...,
    short_help: str | None = ...,
    options_metavar: str | None = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool | str = ...,
    aliases: Optional[Iterable[str]] = ...,
    panels: Optional[List[RichPanel[Any, Any]]] = ...,
) -> Callable[[_AnyCallable], RichGroup]: ...

# variant: with positional name and with positional or keyword cls argument:
# @group(namearg, GroupCls, ...) or @group(namearg, cls=GroupCls, ...)
@overload
def group(
    name: Optional[str],
    cls: Type[G],
    **attrs: Any,
) -> Callable[[_AnyCallable], G]: ...

# variant: name omitted, cls _must_ be a keyword argument, @group(cmd=GroupCls, ...)
@overload
def group(
    name: None = None,
    *,
    cls: Type[G],
    **attrs: Any,
) -> Callable[[_AnyCallable], G]: ...

# variant: with optional string name, no cls argument provided.
@overload
def group(name: Optional[str] = ..., cls: None = None, **attrs: Any) -> Callable[[_AnyCallable], RichGroup]: ...
def group(
    name: Union[str, _AnyCallable, None] = None,
    cls: Optional[Type[G]] = None,
    **attrs: Any,
) -> Union[click.Group, Callable[[_AnyCallable], Union[RichGroup, G]]]: ...
def argument(
    *param_decls: str,
    cls: Optional[Type[click.Argument]] = None,
    type: ParamType | Any | None = None,
    required: bool = False,
    default: Any | Callable[[], Any] | None = None,
    callback: Callable[[click.Context, click.Parameter, Any], Any] | None = None,
    nargs: int | None = None,
    multiple: bool = False,
    metavar: str | None = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: str | Sequence[str] | None = None,
    shell_complete: Callable[[click.Context, click.Parameter, str], list[CompletionItem] | list[str]] | None = None,
    deprecated: bool | str = False,
    panel: Optional[Union[str, List[str]]] = None,
    help: Optional[str] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...
def option(
    *param_decls: str,
    cls: Optional[Type[click.Option]] = None,
    type: Optional[Union["ParamType", Any]] = None,
    required: bool = False,
    default: Optional[Union[Any, Callable[[], Any]]] = None,
    callback: Optional[Callable[[click.Context, click.Parameter, Any], Any]] = None,
    nargs: Optional[int] = None,
    multiple: bool = False,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    shell_complete: Optional[
        Callable[
            [click.Context, click.Parameter, str],
            Union[List[CompletionItem], List[str]],
        ]
    ] = None,
    show_default: Optional[Union[bool, str]] = False,
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    is_flag: Optional[bool] = None,
    flag_value: Optional[Any] = None,
    count: bool = False,
    allow_from_autoenv: bool = True,
    help: Optional[str] = None,
    hidden: bool = False,
    show_choices: bool = True,
    show_envvar: bool = False,
    panel: Optional[Union[str, List[str]]] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...
def password_option(
    *param_decls: str,
    cls: Optional[Type[click.Option]] = None,
    type: Optional[Union["ParamType", Any]] = None,
    required: bool = False,
    default: Optional[Union[Any, Callable[[], Any]]] = None,
    callback: Optional[Callable[[click.Context, click.Parameter, Any], Any]] = None,
    nargs: Optional[int] = None,
    multiple: bool = False,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    shell_complete: Optional[
        Callable[
            [click.Context, click.Parameter, str],
            Union[List[CompletionItem], List[str]],
        ]
    ] = None,
    show_default: Optional[Union[bool, str]] = False,
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    is_flag: Optional[bool] = None,
    flag_value: Optional[Any] = None,
    count: bool = False,
    allow_from_autoenv: bool = True,
    help: Optional[str] = None,
    hidden: bool = False,
    show_choices: bool = True,
    show_envvar: bool = False,
    panel: Optional[Union[str, List[str]]] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...
def version_option(
    version: Optional[str] = None,
    *param_decls: str,
    package_name: Optional[str] = None,
    prog_name: Optional[str] = None,
    message: Optional[str] = None,
    cls: Optional[Type[click.Option]] = None,
    type: Optional[Union["ParamType", Any]] = None,
    required: bool = False,
    default: Optional[Union[Any, Callable[[], Any]]] = None,
    callback: Optional[Callable[[click.Context, click.Parameter, Any], Any]] = None,
    nargs: Optional[int] = None,
    multiple: bool = False,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    shell_complete: Optional[
        Callable[
            [click.Context, click.Parameter, str],
            Union[List[CompletionItem], List[str]],
        ]
    ] = None,
    show_default: Optional[Union[bool, str]] = False,
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    is_flag: Optional[bool] = None,
    flag_value: Optional[Any] = None,
    count: bool = False,
    allow_from_autoenv: bool = True,
    help: Optional[str] = None,
    hidden: bool = False,
    show_choices: bool = True,
    show_envvar: bool = False,
    panel: Optional[Union[str, List[str]]] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...
def help_option(
    *param_decls: str,
    cls: Optional[Type[click.Option]] = None,
    type: Optional[Union["ParamType", Any]] = None,
    required: bool = False,
    default: Optional[Union[Any, Callable[[], Any]]] = None,
    callback: Optional[Callable[[click.Context, click.Parameter, Any], Any]] = None,
    nargs: Optional[int] = None,
    multiple: bool = False,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    shell_complete: Optional[
        Callable[
            [click.Context, click.Parameter, str],
            Union[List[CompletionItem], List[str]],
        ]
    ] = None,
    show_default: Optional[Union[bool, str]] = False,
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    is_flag: Optional[bool] = None,
    flag_value: Optional[Any] = None,
    count: bool = False,
    allow_from_autoenv: bool = True,
    help: Optional[str] = None,
    hidden: bool = False,
    show_choices: bool = True,
    show_envvar: bool = False,
    panel: Optional[Union[str, List[str]]] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...
def confirmation_option(
    *param_decls: str,
    cls: Optional[Type[click.Option]] = None,
    type: Optional[Union["ParamType", Any]] = None,
    required: bool = False,
    default: Optional[Union[Any, Callable[[], Any]]] = None,
    callback: Optional[Callable[[click.Context, click.Parameter, Any], Any]] = None,
    nargs: Optional[int] = None,
    multiple: bool = False,
    metavar: Optional[str] = None,
    expose_value: bool = True,
    is_eager: bool = False,
    envvar: Optional[Union[str, Sequence[str]]] = None,
    shell_complete: Optional[
        Callable[
            [click.Context, click.Parameter, str],
            Union[List[CompletionItem], List[str]],
        ]
    ] = None,
    show_default: Optional[Union[bool, str]] = False,
    prompt: Union[bool, str] = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    hide_input: bool = False,
    is_flag: Optional[bool] = None,
    flag_value: Optional[Any] = None,
    count: bool = False,
    allow_from_autoenv: bool = True,
    help: Optional[str] = None,
    hidden: bool = False,
    show_choices: bool = True,
    show_envvar: bool = False,
    panel: Optional[Union[str, List[str]]] = None,
    help_style: Optional[StyleType] = None,
    **attrs: Any,
) -> Callable[[FC], FC]: ...

# This is a way to trick PyCharm into adding autocompletion for typed dicts
# without jeopardizing anything on Mypy's side.
@overload
def rich_config(
    help_config: RichHelpConfigurationDict,
    *,
    console: Optional[Console] = ...,
) -> Callable[[FC], FC]: ...
@overload
def rich_config(
    help_config: Optional[Union[Dict[str, Any], RichHelpConfigurationDict, RichHelpConfiguration]] = ...,
    *,
    console: Optional[Console] = ...,
) -> Callable[[FC], FC]: ...
def rich_config(
    help_config: Optional[Union[Dict[str, Any], RichHelpConfigurationDict, RichHelpConfiguration]] = None,
    *,
    console: Optional[Console] = None,
) -> Callable[[FC], FC]: ...
@overload
def option_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = ...,
    *,
    options: Optional[List[str]] = ...,
    help: Optional[str] = ...,
    help_style: Optional[StyleType] = ...,
    table_styles: TableKwargs,
    panel_styles: PanelKwargs,
    column_types: Optional[List[OptionColumnType]] = ...,
    inline_help_in_title: Optional[bool] = ...,
    title_style: Optional[StyleType] = ...,
) -> Callable[[FC], FC]: ...
@overload
def option_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = ...,
    *,
    options: Optional[List[str]] = ...,
    help: Optional[str] = ...,
    help_style: Optional[StyleType] = ...,
    table_styles: Union[Dict[str, Any], None] = ...,
    panel_styles: Union[Dict[str, Any], None] = ...,
    column_types: Optional[List[OptionColumnType]] = ...,
    inline_help_in_title: Optional[bool] = ...,
    title_style: Optional[StyleType] = ...,
) -> Callable[[FC], FC]: ...
def option_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = RichOptionPanel,
    *,
    options: Optional[List[str]] = None,
    help: Optional[str] = None,
    help_style: Optional[StyleType] = None,
    table_styles: Union[TableKwargs, Dict[str, Any], None] = None,
    panel_styles: Union[PanelKwargs, Dict[str, Any], None] = None,
    column_types: Optional[List[OptionColumnType]] = None,
    inline_help_in_title: Optional[bool] = None,
    title_style: Optional[StyleType] = None,
) -> Callable[[FC], FC]: ...
@overload
def command_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = RichOptionPanel,
    *,
    commands: Optional[List[str]] = ...,
    help: Optional[str] = ...,
    help_style: Optional[StyleType] = ...,
    table_styles: TableKwargs,
    panel_styles: PanelKwargs,
    column_types: Optional[List[CommandColumnType]] = ...,
    inline_help_in_title: Optional[bool] = ...,
    title_style: Optional[StyleType] = ...,
) -> Callable[[FC], FC]: ...
@overload
def command_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = RichOptionPanel,
    *,
    commands: Optional[List[str]] = ...,
    help: Optional[str] = ...,
    help_style: Optional[StyleType] = ...,
    table_styles: Optional[Dict[str, Any]] = ...,
    panel_styles: Optional[Dict[str, Any]] = ...,
    column_types: Optional[List[CommandColumnType]] = ...,
    inline_help_in_title: Optional[bool] = ...,
    title_style: Optional[StyleType] = ...,
) -> Callable[[FC], FC]: ...
@overload
def command_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = RichOptionPanel,
    *,
    commands: Optional[List[str]] = ...,
    help: Optional[str] = ...,
    help_style: Optional[StyleType] = ...,
    table_styles: None,
    panel_styles: None,
    column_types: Optional[List[CommandColumnType]] = ...,
    inline_help_in_title: Optional[bool] = ...,
    title_style: Optional[StyleType] = ...,
) -> Callable[[FC], FC]: ...
def command_panel(
    name: str,
    cls: Type[RichPanel[click.Parameter]] = RichOptionPanel,
    *,
    commands: Optional[List[str]] = None,
    help: Optional[str] = None,
    help_style: Optional[StyleType] = None,
    table_styles: Optional[Dict[str, Any]] = None,
    panel_styles: Optional[Dict[str, Any]] = None,
    column_types: Optional[List[CommandColumnType]] = None,
    inline_help_in_title: Optional[bool] = None,
    title_style: Optional[StyleType] = None,
) -> Callable[[FC], FC]: ...

PSpec = ParamSpec("PSpec")
R = TypeVar("R")

def pass_context(f: Callable[Concatenate[RichContext, PSpec], R]) -> Callable[PSpec, R]: ...

__all__ = [
    "command",
    "group",
    "argument",
    "option",
    "password_option",
    "confirmation_option",
    "version_option",
    "help_option",
    "rich_config",
    "option_panel",
    "command_panel",
    "pass_context",
]



import pytest
from click import Abort
from click.testing import CliRunner
from inline_snapshot import snapshot

import rich_click
import rich_click.rich_click as rc
from rich_click._compat_click import CLICK_IS_BEFORE_VERSION_821
from rich_click.utils import truthy


@pytest.mark.skipif(CLICK_IS_BEFORE_VERSION_821, reason="CliRunner's stderr capture doesn't work before 8.2.1.")
def test_abort(cli_runner: CliRunner) -> None:
    rc.COLOR_SYSTEM = "truecolor"

    @rich_click.command
    def cli() -> None:
        raise Abort()

    res = cli_runner.invoke(cli)

    assert res.stdout == snapshot("")
    assert res.stderr == snapshot(
        """\
\x1b[31mAborted.\x1b[0m
"""
    )


def test_truthy() -> None:
    assert truthy("true") is True
    assert truthy("1") is True
    assert truthy("false") is False
    assert truthy("0") is False
    assert truthy(None) is None
    assert truthy(10) is True
    assert truthy("a") is None


@pytest.mark.skipif(CLICK_IS_BEFORE_VERSION_821, reason="CliRunner's stderr capture doesn't work before 8.2.1.")
def test_help_to_stderr(cli_runner: CliRunner) -> None:
    @rich_click.command(context_settings={"help_to_stderr": True})
    def cli() -> None:
        """CLI help text"""

    res = cli_runner.invoke(cli, "--help")

    assert res.exit_code == 0
    assert res.stdout == snapshot("")
    assert res.stderr == snapshot(
        """\
                                                                                                    \n\
 Usage: cli [OPTIONS]                                                                               \n\
                                                                                                    \n\
 CLI help text                                                                                      \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, TypedDict


if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

if TYPE_CHECKING:
    from rich.style import StyleType

    from rich_click.rich_help_configuration import CommandColumnType, OptionColumnType


notset: Any = object()


def truthy(o: Any) -> Optional[bool]:
    """Check if string or other obj is truthy."""
    if isinstance(o, str):
        if o.lower() in {"y", "yes", "t", "true", "1"}:
            return True
        elif o.lower() in {"n", "no", "f", "false", "0"}:
            return False
        else:
            return None
    elif o is None:
        return None
    else:
        return bool(o)


def method_is_from_subclass_of(cls: Type[object], base_cls: Type[object], method_name: str) -> bool:
    """
    Check to see whether a class's method comes from a subclass of some base class.

    This is used under the hood to see whether we would expect a patched RichCommand's help text
    methods to be compatible or incompatible with rich-click or not.
    """
    method = getattr(cls, method_name, None)
    if method is None:
        return False
    return any(getattr(c, method_name, None) == method for c in cls.__mro__ if base_cls in c.__mro__)


class CommandGroupDict(TypedDict):
    """Specification for command groups."""

    name: NotRequired[str]
    commands: NotRequired[List[str]]
    help: NotRequired[Optional[str]]
    help_style: NotRequired[Optional["StyleType"]]
    table_styles: NotRequired[Optional[Dict[str, Any]]]
    panel_styles: NotRequired[Optional[Dict[str, Any]]]
    column_types: NotRequired[Optional[List["CommandColumnType"]]]
    inline_help_in_title: NotRequired[Optional[bool]]
    title_style: NotRequired[Optional["StyleType"]]

    deduplicate: NotRequired[bool]


class OptionGroupDict(TypedDict):
    """Specification for option groups."""

    name: NotRequired[str]
    options: NotRequired[List[str]]
    help: NotRequired[Optional[str]]
    help_style: NotRequired[Optional["StyleType"]]
    table_styles: NotRequired[Optional[Dict[str, Any]]]
    panel_styles: NotRequired[Optional[Dict[str, Any]]]
    column_types: NotRequired[Optional[List["OptionColumnType"]]]
    inline_help_in_title: NotRequired[Optional[bool]]
    title_style: NotRequired[Optional["StyleType"]]

    deduplicate: NotRequired[bool]



import importlib
import sys
from pathlib import Path
from typing import Any, Callable, Dict

import pytest

from rich_click.rich_command import RichCommand


if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired


if sys.version_info < (3, 9):
    pytest.skip(reason="typing._eval_type doesn't work how we need it to before 3.9", allow_module_level=True)


def eval_type(t: Any, globalns: Dict[str, Any], localns: Dict[str, Any]) -> Any:
    import typing

    if sys.version_info >= (3, 12):
        return typing._eval_type(t, globalns, localns, type_params=None)  # type: ignore[attr-defined]
    else:
        return typing._eval_type(t, globalns, localns)  # type: ignore[attr-defined]


def test_config_typed_dict_annotations(mock_script_writer: Callable[[str, str], Path]) -> None:
    IGNORED_KEYS = {"highlighter"}

    with open("src/rich_click/_internal_types.py") as f:
        mod_str = f.read()
        mod_str = mod_str.replace("\nif TYPE_CHECKING:", "\nif True:")
        # pyright gets mad at StyleType so we have to replace it with `str`.
        mod_str = mod_str.replace("from rich.style import StyleType", "StyleType = str")
        mock_script_writer(mod_str, "_internal_types_pytest.py")

    # When importing directly from the config module, this test causes side-effects that
    # are very difficult to fully squash.
    #
    # Importing from a copy of the module and editing it in place prevents requiring side-effects
    # which could negatively impact other tests.
    with open("src/rich_click/rich_help_configuration.py") as f:
        mod_str = f.read()
        mod_str = mod_str.replace("\nif TYPE_CHECKING:", "\nif True:")
        # pyright gets mad at StyleType so we have to replace it with `str`.
        mod_str = mod_str.replace("from rich.style import StyleType", "StyleType = str")
        mock_script_writer(mod_str, "_rich_configuration_pytest.py")

    config_mod = importlib.import_module("_rich_configuration_pytest")
    RichHelpConfiguration = config_mod.RichHelpConfiguration

    types_mod = importlib.import_module("_internal_types_pytest")
    RichHelpConfigurationDict = types_mod.RichHelpConfigurationDict

    for attr, field in RichHelpConfiguration.__dataclass_fields__.items():
        if attr in IGNORED_KEYS:
            continue
        expected_ann = field.type
        actual_ann = RichHelpConfigurationDict.__annotations__[attr]

        actual_typ = eval_type(actual_ann, types_mod.__dict__, {})
        expected_typ = eval_type(NotRequired[expected_ann], config_mod.__dict__, {})
        assert actual_typ == expected_typ, f"Failed attr={attr}"

    # Assert the reverse
    assert all(k in RichHelpConfiguration.__dataclass_fields__ for k in RichHelpConfigurationDict.__annotations__)


def test_all_rich_group_attrs_in_annotation(mock_script_writer: Callable[[str, str], Path]) -> None:
    # This is a weak test.
    # It does not assert that annotations, function signatures, etc. are correct.
    # It only asserts that nothing is obviously missing.

    with open("src/rich_click/rich_command.pyi") as f:
        mod_str = f.read()
        mock_script_writer(mod_str, "_rich_command_pyi.py")

    rich_command_pyi_mod = importlib.import_module("_rich_command_pyi")
    AnnotatedRichGroup = rich_command_pyi_mod.RichGroup

    for i in dir(RichCommand):
        assert i in dir(AnnotatedRichGroup)


def test_all_rich_command_attrs_in_annotation(mock_script_writer: Callable[[str, str], Path]) -> None:
    # This is a weak test.
    # It does not assert that annotations, function signatures, etc. are correct.
    # It only asserts that nothing is obviously missing.

    with open("src/rich_click/rich_command.pyi") as f:
        mod_str = f.read()
        mock_script_writer(mod_str, "_rich_command_pyi.py")

    rich_command_pyi_mod = importlib.import_module("_rich_command_pyi")
    AnnotatedRichCommand = rich_command_pyi_mod.RichCommand

    for i in dir(RichCommand):
        assert i in dir(AnnotatedRichCommand)



from __future__ import annotations

import json
import os
from dataclasses import dataclass, field
from types import ModuleType
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from rich_click.rich_click_theme import RichClickTheme, get_theme
from rich_click.utils import CommandGroupDict, OptionGroupDict, notset, truthy


if TYPE_CHECKING:  # pragma: no cover
    from rich.align import AlignMethod
    from rich.box import Box
    from rich.highlighter import Highlighter
    from rich.padding import PaddingDimensions
    from rich.style import StyleType
    from rich.text import Text

T = TypeVar("T", bound="RichHelpConfiguration")

OptionColumnType = Literal[
    "required",
    "opt_primary",
    "opt_secondary",
    "opt_long",
    "opt_short",
    "opt_all",
    "opt_all_metavar",
    "opt_long_metavar",
    "metavar",
    "metavar_short",
    "help",
    # "default",
    # "envvar",
]

CommandColumnType = Literal["name", "aliases", "name_with_aliases", "help"]

OptionHelpSectionType = Literal[
    "help", "required", "envvar", "default", "range", "metavar", "metavar_short", "deprecated"
]

CommandHelpSectionType = Literal["aliases", "help", "deprecated"]

ColumnType = Union[OptionColumnType, CommandColumnType, str]


class FromTheme(object):
    """Sentinel value for unset config options."""

    def __init__(self, default: str) -> None:
        """Initialize a default."""
        self.default = default

    def __repr__(self) -> str:
        return "FromTheme"

    def get_default(self, key: str) -> Any:
        """Get the default value from the default theme."""
        from rich_click.rich_click_theme import get_theme

        theme = get_theme(self.default)
        return theme.styles[key]


FROM_THEME: Any = FromTheme(default="default-box")


def force_terminal_default() -> Optional[bool]:
    """Use as the default factory for `force_terminal`."""
    env_vars = ["FORCE_COLOR", "PY_COLORS", "GITHUB_ACTIONS"]
    for env_var in env_vars:
        if env_var in os.environ:
            return truthy(os.getenv(env_var))
    else:
        return None


def terminal_width_default() -> Optional[int]:
    """Use as the default factory for `width` and `max_width`."""
    width = os.getenv("TERMINAL_WIDTH")
    if width:
        try:
            return int(width)
        except ValueError:
            import warnings

            warnings.warn(
                "Environment variable `TERMINAL_WIDTH` cannot be cast to an integer.", UserWarning, stacklevel=2
            )
            return None
    return None


@dataclass
class RichHelpConfiguration:
    """
    Rich Help Configuration class.

    When merging multiple RichHelpConfigurations together, user-defined values always
    take precedence over the class's defaults. When there are multiple user-defined values
    for a given field, the right-most field is used.
    """

    theme: Optional[Union[str, RichClickTheme]] = field(default=None)
    enable_theme_env_var: bool = field(default=True)

    # Default styles
    style_option: "StyleType" = field(default=FROM_THEME)
    style_option_negative: Optional["StyleType"] = field(default=FROM_THEME)
    style_argument: "StyleType" = field(default=FROM_THEME)
    style_command: "StyleType" = field(default=FROM_THEME)
    style_command_aliases: "StyleType" = field(default=FROM_THEME)
    style_switch: "StyleType" = field(default=FROM_THEME)
    style_switch_negative: Optional["StyleType"] = field(default=FROM_THEME)
    style_metavar: "StyleType" = field(default=FROM_THEME)
    style_metavar_append: "StyleType" = field(default=FROM_THEME)
    style_metavar_separator: "StyleType" = field(default=FROM_THEME)
    style_range_append: "StyleType" = field(default=FROM_THEME)
    style_header_text: "StyleType" = field(default=FROM_THEME)
    style_epilog_text: "StyleType" = field(default=FROM_THEME)
    style_footer_text: "StyleType" = field(default=FROM_THEME)
    style_usage: "StyleType" = field(default=FROM_THEME)
    style_usage_command: "StyleType" = field(default=FROM_THEME)
    style_usage_separator: "StyleType" = field(default=FROM_THEME)
    style_deprecated: "StyleType" = field(default=FROM_THEME)
    style_helptext_first_line: "StyleType" = field(default=FROM_THEME)
    style_helptext: "StyleType" = field(default=FROM_THEME)
    style_helptext_aliases: Optional["StyleType"] = field(default=None)
    style_option_help: "StyleType" = field(default=FROM_THEME)
    style_command_help: "StyleType" = field(default=FROM_THEME)
    style_option_default: "StyleType" = field(default=FROM_THEME)
    style_option_envvar: "StyleType" = field(default=FROM_THEME)
    style_required_short: "StyleType" = field(default=FROM_THEME)
    style_required_long: "StyleType" = field(default=FROM_THEME)
    style_options_panel_border: "StyleType" = field(default=FROM_THEME)
    style_options_panel_box: Optional[Union[str, "Box"]] = field(default=FROM_THEME)
    style_options_panel_help_style: "StyleType" = field(default=FROM_THEME)
    style_options_panel_title_style: "StyleType" = field(default=FROM_THEME)
    style_options_panel_padding: "PaddingDimensions" = field(default=FROM_THEME)
    style_options_panel_style: "StyleType" = field(default=FROM_THEME)
    align_options_panel: "AlignMethod" = field(default="left")
    style_options_table_show_lines: bool = field(default=False)
    style_options_table_leading: int = field(default=0)
    style_options_table_pad_edge: bool = field(default=False)
    style_options_table_padding: "PaddingDimensions" = field(default_factory=lambda: (0, 1))
    style_options_table_expand: bool = field(default=FROM_THEME)
    style_options_table_box: Optional[Union[str, "Box"]] = field(default=FROM_THEME)
    style_options_table_row_styles: Optional[List["StyleType"]] = field(default=None)
    style_options_table_border_style: Optional["StyleType"] = field(default=FROM_THEME)
    style_commands_panel_border: "StyleType" = field(default=FROM_THEME)
    panel_inline_help_in_title: bool = field(default=FROM_THEME)
    panel_inline_help_delimiter: str = field(default=FROM_THEME)
    style_commands_panel_box: Optional[Union[str, "Box"]] = field(default=FROM_THEME)
    style_commands_panel_help_style: "StyleType" = field(default=FROM_THEME)
    style_commands_panel_title_style: "StyleType" = field(default=FROM_THEME)
    style_commands_panel_padding: "PaddingDimensions" = field(default=FROM_THEME)
    style_commands_panel_style: "StyleType" = field(default=FROM_THEME)
    align_commands_panel: "AlignMethod" = field(default="left")
    style_commands_table_show_lines: bool = field(default=False)
    style_commands_table_leading: int = field(default=0)
    style_commands_table_pad_edge: bool = field(default=False)
    style_commands_table_padding: "PaddingDimensions" = field(default_factory=lambda: (0, 1))
    style_commands_table_expand: bool = field(default=FROM_THEME)
    style_commands_table_box: Optional[Union[str, "Box"]] = field(default=FROM_THEME)
    style_commands_table_row_styles: Optional[List["StyleType"]] = field(default=None)
    style_commands_table_border_style: Optional["StyleType"] = field(default=FROM_THEME)
    style_commands_table_column_width_ratio: Optional[Union[Tuple[None, None], Tuple[int, int]]] = field(
        default_factory=lambda: (None, None)
    )
    style_errors_panel_border: "StyleType" = field(default=FROM_THEME)
    style_errors_panel_box: Optional[Union[str, "Box"]] = field(default=FROM_THEME)
    align_errors_panel: "AlignMethod" = field(default="left")
    style_errors_suggestion: Optional["StyleType"] = field(default=None)
    style_errors_suggestion_command: Optional["StyleType"] = field(default=None)
    style_padding_errors: "StyleType" = field(default=FROM_THEME)
    style_aborted: "StyleType" = field(default="red")
    style_padding_usage: "StyleType" = field(default=FROM_THEME)
    style_padding_helptext: "StyleType" = field(default=FROM_THEME)
    style_padding_epilog: "StyleType" = field(default=FROM_THEME)

    panel_title_padding: int = field(default=FROM_THEME)
    width: Optional[int] = field(default_factory=terminal_width_default)
    max_width: Optional[int] = field(default_factory=terminal_width_default)
    color_system: Optional[Literal["auto", "standard", "256", "truecolor", "windows"]] = field(default="auto")
    force_terminal: Optional[bool] = field(default_factory=force_terminal_default)

    options_table_column_types: List[OptionColumnType] = field(default=FROM_THEME)
    commands_table_column_types: List[CommandColumnType] = field(default=FROM_THEME)
    options_table_help_sections: List[OptionHelpSectionType] = field(default=FROM_THEME)
    commands_table_help_sections: List[CommandHelpSectionType] = field(default=FROM_THEME)

    # Fixed strings
    header_text: Optional[Union[str, "Text"]] = field(default=None)
    footer_text: Optional[Union[str, "Text"]] = field(default=None)
    panel_title_string: str = field(default=FROM_THEME)
    deprecated_string: str = field(default=FROM_THEME)
    deprecated_with_reason_string: str = field(default=FROM_THEME)
    default_string: str = field(default=FROM_THEME)
    envvar_string: str = field(default=FROM_THEME)
    required_short_string: str = field(default=FROM_THEME)
    required_long_string: str = field(default=FROM_THEME)
    range_string: str = field(default=FROM_THEME)
    append_metavars_help_string: str = field(default=FROM_THEME)
    append_range_help_string: str = field(default=FROM_THEME)
    helptext_aliases_string: str = field(default="Aliases: {}")
    arguments_panel_title: str = field(default="Arguments")
    options_panel_title: str = field(default="Options")
    commands_panel_title: str = field(default="Commands")
    errors_panel_title: str = field(default="Error")
    delimiter_comma: str = field(default=FROM_THEME)
    delimiter_slash: str = field(default=FROM_THEME)
    errors_suggestion: Optional[Union[str, "Text"]] = field(default=None)
    """Defaults to Try 'cmd -h' for help. Set to False to disable."""
    errors_epilogue: Optional[Union[str, "Text"]] = field(default=None)
    aborted_text: str = field(default="Aborted.")

    padding_header_text: "PaddingDimensions" = field(default=FROM_THEME)
    padding_usage: "PaddingDimensions" = field(default=FROM_THEME)
    padding_helptext: "PaddingDimensions" = field(default=FROM_THEME)
    padding_helptext_deprecated: "PaddingDimensions" = field(default=0)
    padding_helptext_first_line: "PaddingDimensions" = field(default=0)
    padding_epilog: "PaddingDimensions" = field(default=FROM_THEME)
    padding_footer_text: "PaddingDimensions" = field(default=FROM_THEME)
    padding_errors_panel: "PaddingDimensions" = field(default=(0, 0, 1, 0))
    padding_errors_suggestion: "PaddingDimensions" = field(default=(0, 1, 0, 1))
    padding_errors_epilogue: "PaddingDimensions" = field(default=(0, 1, 1, 1))

    # Behaviours
    show_arguments: Optional[bool] = field(default=None)
    """Show positional arguments"""
    show_metavars_column: Optional[bool] = field(default=None)
    """Show a column with the option metavar (eg. INTEGER)"""
    commands_before_options: bool = field(default=False)
    """If set, the commands panel show above the options panel."""
    append_metavars_help: Optional[bool] = field(default=None)
    """Append metavar (eg. [TEXT]) after the help text"""
    group_arguments_options: bool = field(default=False)
    """Show arguments with options instead of in own panel"""
    option_envvar_first: Optional[bool] = field(default=None)
    """Show env vars before option help text instead of after"""
    text_markup: Literal["ansi", "rich", "markdown", None] = field(default=notset)
    """What engine to use to render the text. Default is 'ansi'."""
    text_kwargs: Optional[Dict[str, Any]] = field(default=None)
    """Additional kwargs to pass to Rich text rendering. Kwargs differ by text_markup chosen."""
    text_paragraph_linebreaks: Optional[Literal["\n", "\n\n"]] = field(default=None)
    text_emojis: bool = field(default=notset)
    """If set, parse emoji codes and replace with actual emojis, e.g. :smiley_cat: -> 😺"""
    use_markdown: Optional[bool] = field(default=None)
    """Silently deprecated; use `text_markup` field instead."""
    use_markdown_emoji: Optional[bool] = field(default=None)
    """Silently deprecated; use `text_emojis` instead."""
    use_rich_markup: Optional[bool] = field(default=None)
    """Silently deprecated; use `text_markup` field instead."""
    command_groups: Dict[str, List[CommandGroupDict]] = field(default_factory=lambda: {})
    """Define sorted groups of panels to display subcommands"""
    option_groups: Dict[str, List[OptionGroupDict]] = field(default_factory=lambda: {})
    """Define sorted groups of panels to display options and arguments"""
    use_click_short_help: bool = field(default=False)
    """Use click's default function to truncate help text"""
    helptext_show_aliases: bool = field(default=True)
    highlighter: Optional["Highlighter"] = field(default=None, repr=False, compare=False)
    """(Deprecated) Rich regex highlighter for help highlighting"""

    highlighter_patterns: List[str] = field(
        default_factory=lambda: [
            r"(^|[^\w\-])(?P<switch>-([^\W0-9][\w\-]*\w|[^\W0-9]))",
            r"(^|[^\w\-])(?P<option>--([^\W0-9][\w\-]*\w|[^\W0-9]))",
            r"(?P<metavar><[^>]+>)",
            r"(?P<deprecated>\(DEPRECATED(?:\: .*?)?\))$",
        ]
    )
    """Patterns to use with the option highlighter."""

    legacy_windows: Optional[bool] = field(default=None)

    def __post_init__(self) -> None:  # noqa: D105

        if self.highlighter is not None:
            import warnings

            warnings.warn(
                "`highlighter` kwarg is deprecated in RichHelpConfiguration."
                " Please do one of the following instead: either set highlighter_patterns=[...] if you want"
                " to use regex; or for more advanced use cases where you'd like to use a different type"
                " of rich.highlighter.Highlighter, subclass the `RichHelpFormatter` and update its `highlighter`.",
                DeprecationWarning,
                stacklevel=2,
            )

        if self.use_markdown is not None:
            import warnings

            warnings.warn(
                "`use_markdown=` will be deprecated in a future version of rich-click."
                " Please use `text_markup=` instead.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        if self.use_rich_markup is not None:
            import warnings

            warnings.warn(
                "`use_rich_markup=` will be deprecated in a future version of rich-click."
                " Please use `text_markup=` instead.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        if self.show_metavars_column is not None:
            import warnings

            warnings.warn(
                "`show_metavars_column=` will be deprecated in a future version of rich-click."
                " Please use `options_table_column_types=` instead for rich-click>=1.9.0."
                " The `options_table_column_types` config option lets you specify an ordered list"
                " of which columns are rendered and in what order. The default is:"
                " ['required', 'opt_short', 'opt_long', 'metavar', 'help']."
                " You can remove the metavar column by passing in a new list without 'metavar'.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        if self.append_metavars_help is not None:
            import warnings

            warnings.warn(
                "`append_metavars_help=` will be deprecated in a future version of rich-click."
                " Please use `options_table_help_sections=` instead for rich-click>=1.9.0."
                " The `options_table_help_sections=` config option lets you specify an ordered list"
                " of which sections are rendered and in what order. The default is:"
                " ['help', 'deprecated', 'envvar', 'default', 'required']."
                " You can append the metavar by passing in a new list with 'metavar'.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        if self.option_envvar_first is not None:
            import warnings

            warnings.warn(
                "`option_envvar_first=` will be deprecated in a future version of rich-click."
                " Please use `options_table_help_sections=` instead for rich-click>=1.9.0."
                " The `options_table_help_sections=` config option lets you specify an ordered list"
                " of which sections are rendered and in what order. The default is:"
                " ['help', 'deprecated', 'envvar', 'default', 'required']."
                " You can set the envvar first by passing in a new list with 'envvar' first.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        if self.use_markdown_emoji is not None:
            import warnings

            warnings.warn(
                "`use_markdown_emoji=` will be deprecated in a future version of rich-click."
                " Please use `text_emojis=` instead.",
                PendingDeprecationWarning,
                stacklevel=2,
            )

        self.__dataclass_fields__.pop("highlighter", None)

        self.apply_theme()

    @classmethod
    def load_from_globals(cls, module: Optional[ModuleType] = None, **extra: Any) -> "RichHelpConfiguration":
        """
        Build a RichHelpConfiguration from globals in rich_click.rich_click.

        When building from globals, all fields are treated as having been set by the user,
        meaning they will overwrite other fields when "merged".
        """
        if module is None:
            import rich_click.rich_click as rc

            module = rc
        kw = {}
        for k, v in cls.__dataclass_fields__.items():
            if v.init:
                if k != "highlighter" and hasattr(module, k.upper()):
                    kw[k] = getattr(module, k.upper())

        kw.update(extra)
        inst = cls(**kw)
        return inst

    def apply_theme(self, force_default: bool = False) -> None:
        theme: Optional[Union[str, RichClickTheme]] = None
        raise_key_error = True

        import rich_click.rich_click as rc

        if rc._THEME_FROM_CLI is not None:
            theme = rc._THEME_FROM_CLI
        if self.enable_theme_env_var and "RICH_CLICK_THEME" in os.environ:
            _theme = os.environ["RICH_CLICK_THEME"]
            _theme_cfg = _theme.strip()
            if _theme_cfg.startswith("{"):
                try:
                    data = json.loads(_theme_cfg)
                    # Extract theme before parsing everything else.
                    if isinstance(data, dict) and "theme" in data and theme is None:
                        theme = data["theme"]
                        raise_key_error = False
                        data.pop("theme")
                    for k, v in data.items():
                        if hasattr(self, k):
                            setattr(self, k, v)
                        else:
                            raise TypeError(f"'{type(self)}' has no attribute '{k}'")
                except Exception as e:
                    if force_default:  # only warn once
                        import warnings

                        warnings.warn(
                            f"RICH_CLICK_THEME= failed to parse: {e.__class__.__name__}{e.args}",
                            UserWarning,
                            stacklevel=2,
                        )
            elif theme is None:
                theme = _theme
                raise_key_error = False

        if theme is None:
            theme = self.theme

        theme_styles: Optional[Dict[str, Any]] = None

        if isinstance(theme, RichClickTheme):
            theme_styles = theme.styles
        elif theme is not None:
            theme_styles = get_theme(theme, raise_key_error=raise_key_error).styles

        if theme_styles is not None:
            for k, v in theme_styles.items():
                current = getattr(self, k)
                if isinstance(current, FromTheme):
                    setattr(self, k, v)

        if force_default:
            for k in self.__dataclass_fields__:
                v = getattr(self, k)
                if isinstance(v, FromTheme):
                    setattr(self, k, v.get_default(k))

        if force_default:
            # Handle deprecated fields here
            # must create new copy of these lists; don't modify in-place
            if self.text_markup is notset:
                if self.use_markdown:
                    self.text_markup = "markdown"
                elif self.use_rich_markup:
                    self.text_markup = "rich"
                else:
                    self.text_markup = "ansi"

            if self.text_emojis is notset:
                if self.use_markdown_emoji is not None:
                    self.text_emojis = self.use_markdown_emoji
                elif self.text_emojis is notset:
                    self.text_emojis = self.text_markup in {"markdown", "rich"}

            if self.show_metavars_column is False and "metavar" in self.options_table_column_types:
                self.options_table_column_types = [i for i in self.options_table_column_types if i != "metavar"]

            if self.append_metavars_help is True and "metavar" not in self.options_table_help_sections:
                self.options_table_help_sections = self.options_table_help_sections.copy() + ["metavar"]

            if self.option_envvar_first is True and "envvar" in self.options_table_help_sections:
                self.options_table_help_sections = ["envvar"] + [  # type: ignore[assignment]
                    i for i in self.options_table_help_sections if i != "envvar"
                ]

    def to_theme(self, **kwargs: Any) -> RichClickTheme:
        styles = kwargs.get("styles", {})
        for k, v in self.__dataclass_fields__.items():
            if k == "theme":
                continue
            styles.setdefault(k, getattr(self, k))
        kwargs.setdefault("name", "_from_config")
        kwargs["styles"] = styles
        return RichClickTheme(**kwargs)

    def dump_to_globals(self, module: Optional[ModuleType] = None) -> None:
        if module is None:
            import rich_click.rich_click as rc

            module = rc
        for k, v in self.__dataclass_fields__.items():
            if v.init:
                if hasattr(module, k.upper()):
                    setattr(module, k.upper(), getattr(self, k))


def __getattr__(name: str) -> Any:
    if name == "OptionHighlighter":
        from rich.highlighter import RegexHighlighter

        class OptionHighlighter(RegexHighlighter):
            """Highlights our special options."""

            highlights = [
                r"(^|[^\w\-])(?P<switch>-([^\W0-9][\w\-]*\w|[^\W0-9]))",
                r"(^|[^\w\-])(?P<option>--([^\W0-9][\w\-]*\w|[^\W0-9]))",
                r"(?P<metavar><[^>]+>)",
            ]

        import warnings

        warnings.warn(
            "OptionHighlighter is deprecated and will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )

        globals()["OptionHighlighter"] = OptionHighlighter

        return OptionHighlighter

    else:
        raise AttributeError



from __future__ import annotations

import sys
from gettext import gettext
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type, TypeVar, Union, cast, overload

from click import Argument, Command, Context, Group, Option, Parameter
from click import argument as click_argument
from click import command as click_command
from click import confirmation_option as click_confirmation_option
from click import option as click_option
from click import pass_context as click_pass_context
from click import password_option as click_password_option
from click import version_option as click_version_option

from rich_click.rich_command import RichCommand, RichGroup
from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_panel import RichCommandPanel, RichOptionPanel, RichPanel
from rich_click.rich_parameter import RichArgument, RichOption


if sys.version_info < (3, 10):
    from typing_extensions import Concatenate, ParamSpec
else:
    from typing import Concatenate, ParamSpec


if TYPE_CHECKING:  # pragma: no cover
    from rich.console import Console


_AnyCallable = Callable[..., Any]
F = TypeVar("F", bound=Callable[..., Any])
FC = TypeVar("FC", bound=Union[Command, _AnyCallable])
C = TypeVar("C", bound=Command)


G = TypeVar("G", bound=Group)


# variant: no call, directly as decorator for a function.
@overload
def group(name: _AnyCallable) -> RichGroup: ...


# variant: with positional name and with positional or keyword cls argument:
# @group(namearg, GroupCls, ...) or @group(namearg, cls=GroupCls, ...)
@overload
def group(
    name: Optional[str],
    cls: Type[G],
    **attrs: Any,
) -> Callable[[_AnyCallable], G]: ...


# variant: name omitted, cls _must_ be a keyword argument, @group(cmd=GroupCls, ...)
@overload
def group(
    name: None = None,
    *,
    cls: Type[G],
    **attrs: Any,
) -> Callable[[_AnyCallable], G]: ...


# variant: with optional string name, no cls argument provided.
@overload
def group(name: Optional[str] = ..., cls: None = None, **attrs: Any) -> Callable[[_AnyCallable], RichGroup]: ...


def group(
    name: Union[str, _AnyCallable, None] = None,
    cls: Optional[Type[G]] = None,
    **attrs: Any,
) -> Union[Group, Callable[[_AnyCallable], Union[RichGroup, G]]]:
    """
    Group decorator function.

    Defines the group() function so that it uses the RichGroup class by default.
    """
    if cls is None:
        cls = cast(Type[G], RichGroup)

    if callable(name):
        return command(cls=cls, **attrs)(name)

    return command(name, cls, **attrs)


CmdType = TypeVar("CmdType", bound=Command)


# variant: no call, directly as decorator for a function.
@overload
def command(name: _AnyCallable) -> RichCommand: ...


# variant: with positional name and with positional or keyword cls argument:
# @command(namearg, CommandCls, ...) or @command(namearg, cls=CommandCls, ...)
@overload
def command(
    name: Optional[str],
    cls: Type[CmdType],
    **attrs: Any,
) -> Callable[[_AnyCallable], CmdType]: ...


# variant: name omitted, cls _must_ be a keyword argument, @command(cls=CommandCls, ...)
@overload
def command(
    name: None = None,
    *,
    cls: Type[CmdType],
    **attrs: Any,
) -> Callable[[_AnyCallable], CmdType]: ...


# variant: with optional string name, no cls argument provided.
@overload
def command(name: Optional[str] = ..., cls: None = None, **attrs: Any) -> Callable[[_AnyCallable], RichCommand]: ...


def command(
    name: Union[Optional[str], _AnyCallable] = None,
    cls: Optional[Type[CmdType]] = None,
    **attrs: Any,
) -> Union[Command, Callable[[_AnyCallable], Union[RichCommand, CmdType]]]:
    """
    Command decorator function.

    Defines the command() function so that it uses the RichCommand class by default.
    """
    func = None
    if callable(name):
        func = name
        name = None
        if "__rich_click_cli_patch" not in attrs:
            assert cls is None, "Use 'command(cls=cls)(callable)' to specify a class."
        attrs.pop("__rich_click_cli_patch", None)
        assert not attrs, "Use 'command(**kwargs)(callable)' to provide arguments."
    else:
        attrs.pop("__rich_click_cli_patch", None)

    if cls is None:
        cls = cast(Type[CmdType], RichCommand)

    def decorator(f: _AnyCallable) -> CmdType:
        cs = getattr(f, "__rich_context_settings__", None)
        if cs is not None:
            attr_cs = attrs.pop("context_settings", None)
            attr_cs = attr_cs if attr_cs is not None else {}
            attr_cs.update(cs)
            attrs["context_settings"] = attr_cs
            del f.__rich_context_settings__  # type: ignore[attr-defined]

        panels = getattr(f, "__rich_panels__", None)
        if panels is not None:
            attr_panels = attrs.pop("panels", None)
            attr_panels = attr_panels if attr_panels is not None else []
            attr_panels.extend(reversed(panels))
            attrs["panels"] = attr_panels
            del f.__rich_panels__  # type: ignore[attr-defined]

        return click_command(name, cls, **attrs)(f)

    if func is not None:
        return decorator(func)

    return decorator


def _context_settings_memo(f: Callable[..., Any], extra: Dict[str, Any]) -> None:
    if isinstance(f, RichCommand):
        f.context_settings.update(extra)
    else:
        if not hasattr(f, "__rich_context_settings__"):
            f.__rich_context_settings__ = {}  # type: ignore

        f.__rich_context_settings__.update(extra)  # type: ignore


def _rich_panel_memo(f: Callable[..., Any], panel: RichPanel[Any, Any]) -> None:
    if isinstance(f, RichCommand):
        f.add_panel(panel)
    else:
        if not hasattr(f, "__rich_panels__"):
            f.__rich_panels__ = []  # type: ignore

        f.__rich_panels__.append(panel)  # type: ignore


def rich_config(
    help_config: Optional[Union[Dict[str, Any], RichHelpConfiguration]] = None,
    *,
    console: Optional["Console"] = None,
) -> Callable[[FC], FC]:
    """
    Use decorator to configure Rich Click settings.

    Args:
    ----
        help_config: Rich help configuration that is used internally to format help messages and exceptions
            Defaults to None.
        console: A Rich Console that will be accessible from the `RichContext`, `RichCommand`, and `RichGroup` instances
            Defaults to None.

    """
    from rich.console import Console

    if isinstance(help_config, Console) and console is None:
        import warnings

        warnings.warn(
            "`rich_config()`'s args have been swapped."
            " Please set the config first, and use a kwarg to set the console.",
            DeprecationWarning,
            stacklevel=2,
        )
        console = help_config

    def decorator(obj: FC) -> FC:
        extra: Dict[str, Any] = {}
        if console is not None:
            extra["rich_console"] = console
        if help_config is not None:
            extra["rich_help_config"] = help_config

        _context_settings_memo(obj, extra)

        return obj

    return decorator


def _panel(
    name: str,
    cls: Type[RichPanel[Any, Any]],
    **attrs: Any,
) -> Callable[[FC], FC]:
    def decorator(obj: FC) -> FC:
        _rich_panel_memo(
            obj,
            cls(name=name, **attrs),
        )
        return obj

    return decorator


def option_panel(
    name: str,
    cls: Type[RichPanel[Parameter, Any]] = RichOptionPanel,
    **attrs: Any,
) -> Callable[[FC], FC]:
    """
    Use decorator to create a RichOptionPanel.

    Args:
    ----
        name: Name of the RichOptionPanel instance being created.
        cls: The class of the RichPanel; defaults to RichOptionPanel.
        attrs: Additional attributes to pass to the RichOptionPanel.

    """
    return _panel(name, cls, **attrs)


def command_panel(
    name: str,
    cls: Type[RichPanel[Command, Any]] = RichCommandPanel,
    **attrs: Any,
) -> Callable[[FC], FC]:
    """
    Use decorator to create a RichCommandPanel.

    Args:
    ----
        name: Name of the RichCommandPanel instance being created.
        cls: The class of the RichPanel; defaults to RichCommandPanel.
        attrs: Additional attributes to pass to the RichCommandPanel.

    """
    return _panel(name, cls, **attrs)


# Users of rich_click would face issues using mypy with this code,
# if not for wrapping `pass_context` with a new function signature:
#
# @click.command()
# @click.pass_context
# def cli(ctx: click.RichContext) -> None:
#    ...


P = ParamSpec("P")
R = TypeVar("R")


def pass_context(f: Callable[Concatenate[RichContext, P], R]) -> Callable[P, R]:
    # flake8: noqa: D400,D401
    """Marks a callback as wanting to receive the current context object as first argument."""
    return click_pass_context(f)  # type: ignore[arg-type,unused-ignore]


def help_option(*param_decls: str, **kwargs: Any) -> Callable[[FC], FC]:
    """
    Pre-configured ``--help`` option which immediately prints the help page
    and exits the program.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--help"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def show_help(ctx: Context, param: Parameter, value: bool) -> None:
        """Callback that print the help page on ``<stdout>`` and exits."""
        if value and not ctx.resilient_parsing:
            # Avoid click.echo() because it ignores console settings like force_terminal.
            # Also, do not print() if empty string; assume console was record=False.
            if getattr(ctx, "help_to_stderr", False):
                print(ctx.get_help(), file=sys.stderr)
            else:
                print(ctx.get_help())
            ctx.exit()

    if not param_decls:
        param_decls = ("--help",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", gettext("Show this message and exit."))
    kwargs.setdefault("callback", show_help)
    kwargs.setdefault("cls", RichOption)

    return click_option(*param_decls, **kwargs)


def argument(*param_decls: str, cls: Optional[Type[Argument]] = None, **attrs: Any) -> Callable[[FC], FC]:
    """
    Attaches an argument to the command.  All positional arguments are
    passed as parameter declarations to :class:`Argument`; all keyword
    arguments are forwarded unchanged (except ``cls``).
    This is equivalent to creating an :class:`Argument` instance manually
    and attaching it to the :attr:`Command.params` list.

    For the default argument class, refer to :class:`Argument` and
    :class:`Parameter` for descriptions of parameters.

    :param cls: the argument class to instantiate.  This defaults to
                :class:`Argument`.
    :param param_decls: Passed as positional arguments to the constructor of
        ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """
    if cls is None:
        cls = RichArgument

    return click_argument(*param_decls, cls=cls, **attrs)


def option(*param_decls: str, cls: Optional[Type[Option]] = None, **attrs: Any) -> Callable[[FC], FC]:
    """
    Attaches an option to the command.  All positional arguments are
    passed as parameter declarations to :class:`Option`; all keyword
    arguments are forwarded unchanged (except ``cls``).
    This is equivalent to creating an :class:`Option` instance manually
    and attaching it to the :attr:`Command.params` list.

    For the default option class, refer to :class:`Option` and
    :class:`Parameter` for descriptions of parameters.

    :param cls: the option class to instantiate.  This defaults to
                :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of
        ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """
    if cls is None:
        cls = RichOption

    return click_option(*param_decls, cls=cls, **attrs)


def confirmation_option(*param_decls: str, **kwargs: Any) -> Callable[[FC], FC]:
    """
    Add a ``--yes`` option which shows a prompt before continuing if
    not passed. If the prompt is declined, the program will exit.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--yes"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """
    kwargs.setdefault("cls", RichOption)
    return click_confirmation_option(*param_decls, **kwargs)


def password_option(*param_decls: str, **kwargs: Any) -> Callable[[FC], FC]:
    """
    Add a ``--password`` option which prompts for a password, hiding
    input and asking to enter the value again for confirmation.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--password"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """
    if not param_decls:
        param_decls = ("--password",)

    kwargs.setdefault("prompt", True)
    kwargs.setdefault("confirmation_prompt", True)
    kwargs.setdefault("hide_input", True)
    kwargs.setdefault("cls", RichOption)
    return click_password_option(*param_decls, **kwargs)


def version_option(
    version: Optional[str] = None,
    *param_decls: str,
    package_name: Optional[str] = None,
    prog_name: Optional[str] = None,
    message: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[FC], FC]:
    """
    Add a ``--version`` option which immediately prints the version
    number and exits the program.

    If ``version`` is not provided, Click will try to detect it using
    :func:`importlib.metadata.version` to get the version for the
    ``package_name``. On Python < 3.8, the ``importlib_metadata``
    backport must be installed.

    If ``package_name`` is not provided, Click will try to detect it by
    inspecting the stack frames. This will be used to detect the
    version, so it must match the name of the installed package.

    :param version: The version number to show. If not provided, Click
        will try to detect it.
    :param param_decls: One or more option names. Defaults to the single
        value ``"--version"``.
    :param package_name: The package name to detect the version from. If
        not provided, Click will try to detect it.
    :param prog_name: The name of the CLI to show in the message. If not
        provided, it will be detected from the command.
    :param message: The message to show. The values ``%(prog)s``,
        ``%(package)s``, and ``%(version)s`` are available. Defaults to
        ``"%(prog)s, version %(version)s"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    :raise RuntimeError: ``version`` could not be detected.
    """
    kwargs.setdefault("cls", RichOption)
    return click_version_option(
        version, *param_decls, package_name=package_name, prog_name=prog_name, message=message, **kwargs
    )


__all__ = [
    "command",
    "group",
    "argument",
    "option",
    "password_option",
    "confirmation_option",
    "version_option",
    "help_option",
    "rich_config",
    "option_panel",
    "command_panel",
    "pass_context",
]
